# Copyright 2018-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = (
    "ALL_COMPLETED",
    "FIRST_COMPLETED",
    "FIRST_EXCEPTION",
    "ensure_future",
    "CancelledError",
    "Future",
    "InvalidStateError",
    "TimeoutError",
    "get_child_watcher",
    "get_event_loop",
    "set_child_watcher",
    "get_event_loop_policy",
    "set_event_loop_policy",
    "sleep",
    "Task",
    "wait",
)

import types
import weakref

import asyncio as _real_asyncio

# pylint: disable=redefined-builtin
from asyncio import (
    ALL_COMPLETED,
    CancelledError,
    FIRST_COMPLETED,
    FIRST_EXCEPTION,
    Future,
    InvalidStateError,
    TimeoutError,
)

import threading

import portage

portage.proxy.lazyimport.lazyimport(
    globals(),
    "portage.util.futures.unix_events:_PortageEventLoopPolicy",
    "portage.util.futures:compat_coroutine@_compat_coroutine",
)
from portage.util._eventloop.asyncio_event_loop import (
    AsyncioEventLoop as _AsyncioEventLoop,
)


_lock = threading.Lock()
_policy = None


def get_event_loop_policy():
    """
    Get the current event loop policy.

    @rtype: asyncio.AbstractEventLoopPolicy (or compatible)
    @return: the current event loop policy
    """
    global _lock, _policy
    with _lock:
        if _policy is None:
            _policy = _PortageEventLoopPolicy()
        return _policy


def set_event_loop_policy(policy):
    """
    Set the current event loop policy. If policy is None, the default
    policy is restored.

    @type policy: asyncio.AbstractEventLoopPolicy or None
    @param policy: new event loop policy
    """
    global _lock, _policy
    with _lock:
        _policy = policy or _PortageEventLoopPolicy()


def get_event_loop():
    """
    Equivalent to calling get_event_loop_policy().get_event_loop().

    @rtype: asyncio.AbstractEventLoop (or compatible)
    @return: the event loop for the current context
    """
    return get_event_loop_policy().get_event_loop()


def get_child_watcher():
    """Equivalent to calling get_event_loop_policy().get_child_watcher()."""
    return get_event_loop_policy().get_child_watcher()


def set_child_watcher(watcher):
    """Equivalent to calling
    get_event_loop_policy().set_child_watcher(watcher)."""
    return get_event_loop_policy().set_child_watcher(watcher)


def create_subprocess_exec(*args, **kwargs):
    """
    Create a subprocess.

    @param args: program and arguments
    @type args: str
    @param stdin: stdin file descriptor
    @type stdin: file or int
    @param stdout: stdout file descriptor
    @type stdout: file or int
    @param stderr: stderr file descriptor
    @type stderr: file or int
    @param close_fds: close file descriptors
    @type close_fds: bool
    @param loop: asyncio.AbstractEventLoop (or compatible)
    @type loop: event loop
    @type kwargs: varies
    @param kwargs: subprocess.Popen parameters
    @rtype: asyncio.subprocess.Process (or compatible)
    @return: asyncio.subprocess.Process interface
    """
    loop = _wrap_loop(kwargs.pop("loop", None))
    # Python 3.4 and later implement PEP 446, which makes newly
    # created file descriptors non-inheritable by default.
    kwargs.setdefault("close_fds", False)
    # Use the real asyncio create_subprocess_exec (loop argument
    # is deprecated since since Python 3.8).
    return ensure_future(
        _real_asyncio.create_subprocess_exec(*args, **kwargs), loop=loop
    )


def wait(futures, loop=None, timeout=None, return_when=ALL_COMPLETED):
    """
    Wraps asyncio.wait() and omits the loop argument which is not
    supported since python 3.10.
    """
    return _real_asyncio.wait(futures, timeout=timeout, return_when=return_when)


def iscoroutinefunction(func):
    """
    Return True if func is a decorated coroutine function,
    supporting both asyncio.coroutine and compat_coroutine since
    their behavior is identical for all practical purposes.
    """
    if _compat_coroutine._iscoroutinefunction(func):
        return True
    if _real_asyncio.iscoroutinefunction(func):
        return True
    return False


class Task(Future):
    """
    Schedule the execution of a coroutine: wrap it in a future. A task
    is a subclass of Future.
    """

    def __init__(self, coro, loop=None):
        raise NotImplementedError


def ensure_future(coro_or_future, loop=None):
    """
    Wrap a coroutine or an awaitable in a future.

    If the argument is a Future, it is returned directly.

    @type coro_or_future: coroutine or Future
    @param coro_or_future: coroutine or future to wrap
    @type loop: asyncio.AbstractEventLoop (or compatible)
    @param loop: event loop
    @rtype: asyncio.Future (or compatible)
    @return: an instance of Future
    """
    loop = _wrap_loop(loop)
    if isinstance(loop._asyncio_wrapper, _AsyncioEventLoop):
        # Use the real asyncio loop and ensure_future.
        return _real_asyncio.ensure_future(
            coro_or_future, loop=loop._asyncio_wrapper._loop
        )

    if isinstance(coro_or_future, Future):
        return coro_or_future
    raise NotImplementedError


def sleep(delay, result=None, loop=None):
    """
    Create a future that completes after a given time (in seconds). If
    result is provided, it is produced to the caller when the future
    completes.

    @type delay: int or float
    @param delay: delay seconds
    @type result: object
    @param result: result of the future
    @type loop: asyncio.AbstractEventLoop (or compatible)
    @param loop: event loop
    @rtype: asyncio.Future (or compatible)
    @return: an instance of Future
    """
    loop = _wrap_loop(loop)
    future = loop.create_future()
    handle = loop.call_later(delay, future.set_result, result)

    def cancel_callback(future):
        if future.cancelled():
            handle.cancel()

    future.add_done_callback(cancel_callback)
    return future


def _wrap_loop(loop=None):
    """
    In order to deal with asyncio event loop compatibility issues,
    use this function to wrap the loop parameter for functions
    that support it. For example, since python3.4 does not have the
    AbstractEventLoop.create_future() method, this helper function
    can be used to add a wrapper that implements the create_future
    method for python3.4.

    @type loop: asyncio.AbstractEventLoop (or compatible)
    @param loop: event loop
    @rtype: asyncio.AbstractEventLoop (or compatible)
    @return: event loop
    """
    # The default loop returned by _wrap_loop should be consistent
    # with global_event_loop, in order to avoid accidental registration
    # of callbacks with a loop that is not intended to run.
    loop = loop or _safe_loop()
    return loop if hasattr(loop, "_asyncio_wrapper") else _AsyncioEventLoop(loop=loop)


def _safe_loop():
    """
    Return an event loop that's safe to use within the current context.
    For portage internal callers or external API consumers calling from
    the main thread, this returns a globally shared event loop instance.

    For external API consumers calling from a non-main thread, an
    asyncio loop must be registered for the current thread, or else the
    asyncio.get_event_loop() function will raise an error like this:

      RuntimeError: There is no current event loop in thread 'Thread-1'.

    In order to avoid this RuntimeError, a loop will be automatically
    created like this:

      asyncio.set_event_loop(asyncio.new_event_loop())

    In order to avoid a ResourceWarning, automatically created loops
    are added to a WeakValueDictionary, and closed via an atexit hook
    if they still exist during exit for the current pid.

    @rtype: asyncio.AbstractEventLoop (or compatible)
    @return: event loop instance
    """
    loop = _get_running_loop()
    if loop is not None:
        return loop

    thread_key = threading.get_ident()
    with _thread_weakrefs.lock:
        if _thread_weakrefs.pid != portage.getpid():
            _thread_weakrefs.pid = portage.getpid()
            _thread_weakrefs.mainloop = None
            _thread_weakrefs.loops = weakref.WeakValueDictionary()
        try:
            loop = _thread_weakrefs.loops[thread_key]
        except KeyError:
            try:
                try:
                    _loop = _real_asyncio.get_running_loop()
                except AttributeError:
                    _loop = _real_asyncio.get_event_loop()
            except RuntimeError:
                _loop = _real_asyncio.new_event_loop()
                _real_asyncio.set_event_loop(_loop)
            loop = _thread_weakrefs.loops[thread_key] = _AsyncioEventLoop(loop=_loop)

    if (
        _thread_weakrefs.mainloop is None
        and threading.current_thread() is threading.main_thread()
    ):
        _thread_weakrefs.mainloop = loop

    return loop


def _get_running_loop():
    with _thread_weakrefs.lock:
        if _thread_weakrefs.pid == portage.getpid():
            try:
                loop = _thread_weakrefs.loops[threading.get_ident()]
            except KeyError:
                return None
            return loop if loop.is_running() else None


def _thread_weakrefs_atexit():
    with _thread_weakrefs.lock:
        if _thread_weakrefs.pid == portage.getpid():
            while True:
                try:
                    thread_key, loop = _thread_weakrefs.loops.popitem()
                except KeyError:
                    break
                else:
                    loop.close()


_thread_weakrefs = types.SimpleNamespace(
    lock=threading.Lock(), loops=None, mainloop=None, pid=None
)
portage.process.atexit_register(_thread_weakrefs_atexit)



# Copyright 2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage import os
from portage.checksum import perform_multiple_checksums
from portage.util._async.ForkProcess import ForkProcess
from _emerge.PipeReader import PipeReader


class FileDigester(ForkProcess):
    """
    Asynchronously generate file digests. Pass in file_path and
    hash_names, and after successful execution, the digests
    attribute will be a dict containing all of the requested
    digests.
    """

    __slots__ = (
        "file_path",
        "digests",
        "hash_names",
        "_digest_pipe_reader",
        "_digest_pw",
    )

    def _start(self):
        pr, pw = os.pipe()
        self.fd_pipes = {}
        self.fd_pipes[pw] = pw
        self._digest_pw = pw
        self._digest_pipe_reader = PipeReader(
            input_files={"input": pr}, scheduler=self.scheduler
        )
        self._digest_pipe_reader.addExitListener(self._digest_pipe_reader_exit)
        self._digest_pipe_reader.start()
        ForkProcess._start(self)
        os.close(pw)

    def _run(self):
        digests = perform_multiple_checksums(self.file_path, hashes=self.hash_names)

        buf = "".join("%s=%s\n" % item for item in digests.items()).encode("utf_8")

        while buf:
            buf = buf[os.write(self._digest_pw, buf) :]

        return os.EX_OK

    def _parse_digests(self, data):
        digests = {}
        for line in data.decode("utf_8").splitlines():
            parts = line.split("=", 1)
            if len(parts) == 2:
                digests[parts[0]] = parts[1]

        self.digests = digests

    def _async_waitpid(self):
        # Ignore this event, since we want to ensure that we
        # exit only after _digest_pipe_reader has reached EOF.
        if self._digest_pipe_reader is None:
            ForkProcess._async_waitpid(self)

    def _digest_pipe_reader_exit(self, pipe_reader):
        self._parse_digests(pipe_reader.getvalue())
        self._digest_pipe_reader = None
        if self.pid is None:
            self._unregister()
            self._async_wait()
        else:
            self._async_waitpid()

    def _unregister(self):
        ForkProcess._unregister(self)

        pipe_reader = self._digest_pipe_reader
        if pipe_reader is not None:
            self._digest_pipe_reader = None
            pipe_reader.removeExitListener(self._digest_pipe_reader_exit)
            pipe_reader.cancel()



# Copyright 2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import portage
from portage import os
from portage.exception import FileNotFound, PermissionDenied, PortagePackageException
from portage.localization import _
from portage.util._async.ForkProcess import ForkProcess


class ManifestProcess(ForkProcess):
    __slots__ = ("cp", "distdir", "fetchlist_dict", "repo_config")

    MODIFIED = 16

    def _run(self):
        mf = self.repo_config.load_manifest(
            os.path.join(self.repo_config.location, self.cp),
            self.distdir,
            fetchlist_dict=self.fetchlist_dict,
        )

        try:
            mf.create(assumeDistHashesAlways=True)
        except FileNotFound as e:
            portage.writemsg(
                _("!!! File %s doesn't exist, can't update " "Manifest\n") % e,
                noiselevel=-1,
            )
            return 1

        except PortagePackageException as e:
            portage.writemsg(f"!!! {e}\n", noiselevel=-1)
            return 1

        try:
            modified = mf.write(sign=False)
        except PermissionDenied as e:
            portage.writemsg(
                f"!!! {_('Permission Denied')}: {e}\n",
                noiselevel=-1,
            )
            return 1
        else:
            if modified:
                return self.MODIFIED
            return os.EX_OK



# xpak/test_decodeint.py
# Copyright 2006-2020 Gentoo Authors
# Portage Unit Testing Functionality

from portage.tests import TestCase
from portage.xpak import decodeint, encodeint


class testDecodeIntTestCase(TestCase):
    def testDecodeInt(self):
        for n in range(1000):
            self.assertEqual(decodeint(encodeint(n)), n)

        for n in (2**32 - 1,):
            self.assertEqual(decodeint(encodeint(n)), n)



# Copyright 2017-2019 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class SlotOperatorRuntimePkgMaskTestCase(TestCase):
    def testSlotOperatorRuntimePkgMask(self):
        ebuilds = {
            "app-misc/meta-pkg-2": {
                "EAPI": "6",
                "DEPEND": "=app-misc/B-2 =app-misc/C-1  =app-misc/D-1 =dev-libs/foo-2",
                "RDEPEND": "=app-misc/B-2 =app-misc/C-1 =app-misc/D-1 =dev-libs/foo-2",
            },
            "app-misc/meta-pkg-1": {
                "EAPI": "6",
                "DEPEND": "=app-misc/B-1 =app-misc/C-1  =app-misc/D-1 =dev-libs/foo-1",
                "RDEPEND": "=app-misc/B-1 =app-misc/C-1 =app-misc/D-1 =dev-libs/foo-1",
            },
            "app-misc/B-1": {
                "EAPI": "6",
                "DEPEND": "dev-libs/foo:=",
                "RDEPEND": "dev-libs/foo:=",
            },
            "app-misc/B-2": {
                "EAPI": "6",
                "DEPEND": "dev-libs/foo:=",
                "RDEPEND": "dev-libs/foo:=",
            },
            "app-misc/C-1": {
                "EAPI": "6",
                "DEPEND": "dev-libs/foo:=",
                "RDEPEND": "dev-libs/foo:=",
            },
            "app-misc/C-2": {
                "EAPI": "6",
                "DEPEND": "dev-libs/foo:=",
                "RDEPEND": "dev-libs/foo:=",
            },
            "app-misc/D-1": {
                "EAPI": "6",
                "DEPEND": "dev-libs/foo:=",
                "RDEPEND": "dev-libs/foo:=",
            },
            "app-misc/D-2": {
                "EAPI": "6",
                "DEPEND": "dev-libs/foo:=",
                "RDEPEND": "dev-libs/foo:=",
            },
            "dev-libs/foo-1": {
                "EAPI": "6",
                "SLOT": "0/1",
            },
            "dev-libs/foo-2": {
                "EAPI": "6",
                "SLOT": "0/2",
            },
        }

        installed = {
            "app-misc/meta-pkg-1": {
                "EAPI": "6",
                "DEPEND": "=app-misc/B-1 =app-misc/C-1  =app-misc/D-1 =dev-libs/foo-1",
                "RDEPEND": "=app-misc/B-1 =app-misc/C-1 =app-misc/D-1 =dev-libs/foo-1",
            },
            "app-misc/B-1": {
                "EAPI": "6",
                "DEPEND": "dev-libs/foo:0/1=",
                "RDEPEND": "dev-libs/foo:0/1=",
            },
            "app-misc/C-1": {
                "EAPI": "6",
                "DEPEND": "dev-libs/foo:0/1=",
                "RDEPEND": "dev-libs/foo:0/1=",
            },
            "app-misc/D-1": {
                "EAPI": "6",
                "DEPEND": "dev-libs/foo:0/1=",
                "RDEPEND": "dev-libs/foo:0/1=",
            },
            "dev-libs/foo-1": {
                "EAPI": "6",
                "SLOT": "0/1",
            },
        }

        world = ("app-misc/meta-pkg",)

        test_cases = (
            ResolverPlaygroundTestCase(
                ["=app-misc/meta-pkg-2"],
                options={
                    "--backtrack": 14,
                },
                success=True,
                ambiguous_merge_order=True,
                mergelist=[
                    "dev-libs/foo-2",
                    ("app-misc/D-1", "app-misc/C-1", "app-misc/B-2"),
                    "app-misc/meta-pkg-2",
                ],
            ),
        )

        playground = ResolverPlayground(
            debug=False, ebuilds=ebuilds, installed=installed, world=world
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            # Disable debug so that cleanup works.
            playground.debug = False
            playground.cleanup()



# Copyright 1999-2019 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import errno
import io
import sys

from _emerge.CompositeTask import CompositeTask
import portage
from portage import os
from portage.checksum import (
    _apply_hash_filter,
    _filter_unaccelarated_hashes,
    _hash_filter,
)
from portage.output import EOutput
from portage.util._async.FileDigester import FileDigester
from portage.package.ebuild.fetch import _checksum_failure_temp_file


class BinpkgVerifier(CompositeTask):
    __slots__ = ("logfile", "pkg", "_digests", "_pkg_path")

    def _start(self):
        bintree = self.pkg.root_config.trees["bintree"]
        digests = bintree._get_digests(self.pkg)
        if "size" not in digests:
            self.returncode = os.EX_OK
            self._async_wait()
            return

        digests = _filter_unaccelarated_hashes(digests)
        hash_filter = _hash_filter(bintree.settings.get("PORTAGE_CHECKSUM_FILTER", ""))
        if not hash_filter.transparent:
            digests = _apply_hash_filter(digests, hash_filter)

        self._digests = digests

        try:
            size = os.stat(self._pkg_path).st_size
        except OSError as e:
            if e.errno not in (errno.ENOENT, errno.ESTALE):
                raise
            self.scheduler.output(
                f"!!! Fetching Binary failed for '{self.pkg.cpv}'\n",
                log_path=self.logfile,
                background=self.background,
            )
            self.returncode = 1
            self._async_wait()
            return
        else:
            if size != digests["size"]:
                self._digest_exception("size", size, digests["size"])
                self.returncode = 1
                self._async_wait()
                return

        self._start_task(
            FileDigester(
                file_path=self._pkg_path,
                hash_names=(k for k in digests if k != "size"),
                background=self.background,
                logfile=self.logfile,
                scheduler=self.scheduler,
            ),
            self._digester_exit,
        )

    def _digester_exit(self, digester):
        if self._default_exit(digester) != os.EX_OK:
            self.wait()
            return

        for hash_name in digester.hash_names:
            if digester.digests[hash_name] != self._digests[hash_name]:
                self._digest_exception(
                    hash_name, digester.digests[hash_name], self._digests[hash_name]
                )
                self.returncode = 1
                self.wait()
                return

        if self.pkg.root_config.settings.get("PORTAGE_QUIET") != "1":
            self._display_success()

        self.returncode = os.EX_OK
        self.wait()

    def _display_success(self):
        stdout_orig = sys.stdout
        stderr_orig = sys.stderr
        global_havecolor = portage.output.havecolor
        out = io.StringIO()
        try:
            sys.stdout = out
            sys.stderr = out
            if portage.output.havecolor:
                portage.output.havecolor = not self.background

            path = self._pkg_path
            if path.endswith(".partial"):
                path = path[: -len(".partial")]
            eout = EOutput()
            eout.ebegin(
                f"{os.path.basename(path)} {' '.join(sorted(self._digests))} ;-)"
            )
            eout.eend(0)

        finally:
            sys.stdout = stdout_orig
            sys.stderr = stderr_orig
            portage.output.havecolor = global_havecolor

        self.scheduler.output(
            out.getvalue(), log_path=self.logfile, background=self.background
        )

    def _digest_exception(self, name, value, expected):
        head, tail = os.path.split(self._pkg_path)
        temp_filename = _checksum_failure_temp_file(
            self.pkg.root_config.settings, head, tail
        )

        self.scheduler.output(
            "\n!!! Digest verification failed:\n"
            f"!!! {self._pkg_path}\n"
            f"!!! Reason: Failed on {name} verification\n"
            f"!!! Got: {value}\n"
            f"!!! Expected: {expected}\n"
            f"File renamed to '{temp_filename}'\n",
            log_path=self.logfile,
            background=self.background,
        )



# Copyright 2011-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import pytest

from portage.tests import TestCase
from portage.util.whirlpool import CWhirlpool, PyWhirlpool


CWHIRLPOOL_AVAILABLE = CWhirlpool.is_available
CWHIRLPOOL_NOT_AVAILABLE_MSG = "C Whirlpool extension is not importable"


class WhirlpoolTestCase(TestCase):
    def testBundledWhirlpool(self, cls=PyWhirlpool):
        self.assertEqual(
            cls(b"The quick brown fox jumps over the lazy dog").hexdigest(),
            "b97de512e91e3828b40d2b0fdce9ceb3c4a71f9bea8d88e75c4fa854df36725fd2b52eb6544edcacd6f8beddfea403cb55ae31f03ad62a5ef54e42ee82c3fb35",
        )
        self.assertEqual(
            cls(b"The quick brown fox jumps over the lazy eog").hexdigest(),
            "c27ba124205f72e6847f3e19834f925cc666d0974167af915bb462420ed40cc50900d85a1f923219d832357750492d5c143011a76988344c2635e69d06f2d38c",
        )
        self.assertEqual(
            cls(b"").hexdigest(),
            "19fa61d75522a4669b44e39c1d2e1726c530232130d407f89afee0964997f7a73e83be698b288febcf88e3e03c4f0757ea8964e59b63d93708b138cc42a66eb3",
        )
        w = cls()
        w.update(b"")
        self.assertEqual(
            w.hexdigest(),
            "19fa61d75522a4669b44e39c1d2e1726c530232130d407f89afee0964997f7a73e83be698b288febcf88e3e03c4f0757ea8964e59b63d93708b138cc42a66eb3",
        )

    @pytest.mark.skipif(
        not CWHIRLPOOL_AVAILABLE,
        reason=CWHIRLPOOL_NOT_AVAILABLE_MSG,
    )
    def testCWhirlpool(self):
        if not CWHIRLPOOL_AVAILABLE:
            self.skipTest(CWHIRLPOOL_NOT_AVAILABLE_MSG)
        self.testBundledWhirlpool(CWhirlpool)



# Copyright: 2005-2020 Gentoo Authors
# Author(s): Brian Harring (ferringb@gentoo.org)
# License: GPL2

from portage.cache import template, cache_errors
from portage.cache.template import reconstruct_eclasses


class SQLDatabase(template.database):
    """template class for RDBM based caches

    This class is designed such that derivatives don't have to change much code, mostly constant strings.
    _BaseError must be an exception class that all Exceptions thrown from the derived RDBMS are derived
    from.

    SCHEMA_INSERT_CPV_INTO_PACKAGE should be modified dependant on the RDBMS, as should SCHEMA_PACKAGE_CREATE-
    basically you need to deal with creation of a unique pkgid.  If the dbapi2 rdbms class has a method of
    recovering that id, then modify _insert_cpv to remove the extra select.

    Creation of a derived class involves supplying _initdb_con, and table_exists.
    Additionally, the default schemas may have to be modified.
    """

    SCHEMA_PACKAGE_NAME = "package_cache"
    SCHEMA_PACKAGE_CREATE = (
        "CREATE TABLE %s (\
		pkgid INTEGER PRIMARY KEY, label VARCHAR(255), cpv VARCHAR(255), UNIQUE(label, cpv))"
        % SCHEMA_PACKAGE_NAME
    )
    SCHEMA_PACKAGE_DROP = f"DROP TABLE {SCHEMA_PACKAGE_NAME}"

    SCHEMA_VALUES_NAME = "values_cache"
    SCHEMA_VALUES_CREATE = (
        "CREATE TABLE %s ( pkgid integer references %s (pkgid) on delete cascade, \
		key varchar(255), value text, UNIQUE(pkgid, key))"
        % (SCHEMA_VALUES_NAME, SCHEMA_PACKAGE_NAME)
    )
    SCHEMA_VALUES_DROP = f"DROP TABLE {SCHEMA_VALUES_NAME}"
    SCHEMA_INSERT_CPV_INTO_PACKAGE = (
        "INSERT INTO %s (label, cpv) VALUES(%%s, %%s)" % SCHEMA_PACKAGE_NAME
    )

    _BaseError = ()
    _dbClass = None

    autocommits = False
    # 	cleanse_keys = True

    # boolean indicating if the derived RDBMS class supports replace syntax
    _supports_replace = False

    def __init__(self, location, label, auxdbkeys, *args, **config):
        """initialize the instance.
        derived classes shouldn't need to override this"""

        super().__init__(location, label, auxdbkeys, *args, **config)

        config.setdefault("host", "127.0.0.1")
        config.setdefault("autocommit", self.autocommits)
        self._initdb_con(config)

        self.label = self._sfilter(self.label)

    def _dbconnect(self, config):
        """should be overridden if the derived class needs special parameters for initializing
        the db connection, or cursor"""
        self.db = self._dbClass(**config)
        self.con = self.db.cursor()

    def _initdb_con(self, config):
        """ensure needed tables are in place.
        If the derived class needs a different set of table creation commands, overload the approriate
        SCHEMA_ attributes.  If it needs additional execution beyond, override"""

        self._dbconnect(config)
        if not self._table_exists(self.SCHEMA_PACKAGE_NAME):
            if self.readonly:
                raise cache_errors.ReadOnlyRestriction(
                    f"table {self.SCHEMA_PACKAGE_NAME} doesn't exist"
                )
            try:
                self.con.execute(self.SCHEMA_PACKAGE_CREATE)
            except self._BaseError as e:
                raise cache_errors.InitializationError(self.__class__, e)

        if not self._table_exists(self.SCHEMA_VALUES_NAME):
            if self.readonly:
                raise cache_errors.ReadOnlyRestriction(
                    f"table {self.SCHEMA_VALUES_NAME} doesn't exist"
                )
            try:
                self.con.execute(self.SCHEMA_VALUES_CREATE)
            except self._BaseError as e:
                raise cache_errors.InitializationError(self.__class__, e)

    def _table_exists(self, tbl):
        """return true if a table exists
        derived classes must override this"""
        raise NotImplementedError

    def _sfilter(self, s):
        """meta escaping, returns quoted string for use in sql statements"""
        return '"%s"' % s.replace("\\", "\\\\").replace('"', '\\"')

    def _getitem(self, cpv):
        try:
            self.con.execute(
                "SELECT key, value FROM %s NATURAL JOIN %s "
                "WHERE label=%s AND cpv=%s"
                % (
                    self.SCHEMA_PACKAGE_NAME,
                    self.SCHEMA_VALUES_NAME,
                    self.label,
                    self._sfilter(cpv),
                )
            )
        except self._BaseError as e:
            raise cache_errors.CacheCorruption(self, cpv, e)

        rows = self.con.fetchall()

        if len(rows) == 0:
            raise KeyError(cpv)

        vals = {k: "" for k in self._known_keys}
        vals.update(dict(rows))
        return vals

    def _delitem(self, cpv):
        """delete a cpv cache entry
        derived RDBM classes for this *must* either support cascaded deletes, or
        override this method"""
        try:
            try:
                self.con.execute(
                    "DELETE FROM %s WHERE label=%s AND cpv=%s"
                    % (self.SCHEMA_PACKAGE_NAME, self.label, self._sfilter(cpv))
                )
                if self.autocommits:
                    self.commit()
            except self._BaseError as e:
                raise cache_errors.CacheCorruption(self, cpv, e)
            if self.con.rowcount <= 0:
                raise KeyError(cpv)
        except SystemExit:
            raise
        except Exception:
            if not self.autocommits:
                self.db.rollback()
                # yes, this can roll back a lot more then just the delete.  deal.
            raise

    def __del__(self):
        # just to be safe.
        if "db" in self.__dict__ and self.db is not None:
            self.commit()
            self.db.close()

    def _setitem(self, cpv, values):
        try:
            # insert.
            try:
                pkgid = self._insert_cpv(cpv)
            except self._BaseError as e:
                raise cache_errors.CacheCorruption(cpv, e)

            # __getitem__ fills out missing values,
            # so we store only what's handed to us and is a known key
            db_values = []
            for key in self._known_keys:
                if key in values and values[key]:
                    db_values.append({"key": key, "value": values[key]})

            if len(db_values) > 0:
                try:
                    self.con.executemany(
                        'INSERT INTO %s (pkgid, key, value) VALUES("%s", %%(key)s, %%(value)s)'
                        % (self.SCHEMA_VALUES_NAME, str(pkgid)),
                        db_values,
                    )
                except self._BaseError as e:
                    raise cache_errors.CacheCorruption(cpv, e)
            if self.autocommits:
                self.commit()

        except SystemExit:
            raise
        except Exception:
            if not self.autocommits:
                try:
                    self.db.rollback()
                except self._BaseError:
                    pass
            raise

    def _insert_cpv(self, cpv):
        """uses SCHEMA_INSERT_CPV_INTO_PACKAGE, which must be overloaded if the table definition
        doesn't support auto-increment columns for pkgid.
        returns the cpvs new pkgid
        note this doesn't commit the transaction.  The caller is expected to."""

        cpv = self._sfilter(cpv)
        if self._supports_replace:
            query_str = self.SCHEMA_INSERT_CPV_INTO_PACKAGE.replace(
                "INSERT", "REPLACE", 1
            )
        else:
            # just delete it.
            try:
                del self[cpv]
            except (cache_errors.CacheCorruption, KeyError):
                pass
            query_str = self.SCHEMA_INSERT_CPV_INTO_PACKAGE
        try:
            self.con.execute(query_str % (self.label, cpv))
        except self._BaseError:
            self.db.rollback()
            raise
        self.con.execute(
            "SELECT pkgid FROM %s WHERE label=%s AND cpv=%s"
            % (self.SCHEMA_PACKAGE_NAME, self.label, cpv)
        )

        if self.con.rowcount != 1:
            raise cache_error.CacheCorruption(
                cpv,
                "Tried to insert the cpv, but found "
                " %i matches upon the following select!" % len(rows),
            )
        return self.con.fetchone()[0]

    def __contains__(self, cpv):
        if not self.autocommits:
            try:
                self.commit()
            except self._BaseError as e:
                raise cache_errors.GeneralCacheCorruption(e)

        try:
            self.con.execute(
                "SELECT cpv FROM %s WHERE label=%s AND cpv=%s"
                % (self.SCHEMA_PACKAGE_NAME, self.label, self._sfilter(cpv))
            )
        except self._BaseError as e:
            raise cache_errors.GeneralCacheCorruption(e)
        return self.con.rowcount > 0

    def __iter__(self):
        if not self.autocommits:
            try:
                self.commit()
            except self._BaseError as e:
                raise cache_errors.GeneralCacheCorruption(e)

        try:
            self.con.execute(
                f"SELECT cpv FROM {self.SCHEMA_PACKAGE_NAME} WHERE label={self.label}"
            )
        except self._BaseError as e:
            raise cache_errors.GeneralCacheCorruption(e)
        # 		return [ row[0] for row in self.con.fetchall() ]
        for x in self.con.fetchall():
            yield x[0]

    def iteritems(self):
        try:
            self.con.execute(
                "SELECT cpv, key, value FROM %s NATURAL JOIN %s "
                "WHERE label=%s"
                % (self.SCHEMA_PACKAGE_NAME, self.SCHEMA_VALUES_NAME, self.label)
            )
        except self._BaseError as e:
            raise cache_errors.CacheCorruption(self, cpv, e)

        oldcpv = None
        l = []
        for x, y, v in self.con.fetchall():
            if oldcpv != x:
                if oldcpv is not None:
                    d = dict(l)
                    if "_eclasses_" in d:
                        d["_eclasses_"] = reconstruct_eclasses(oldcpv, d["_eclasses_"])
                    else:
                        d["_eclasses_"] = {}
                    yield cpv, d
                l.clear()
                oldcpv = x
            l.append((y, v))
        if oldcpv is not None:
            d = dict(l)
            if "_eclasses_" in d:
                d["_eclasses_"] = reconstruct_eclasses(oldcpv, d["_eclasses_"])
            else:
                d["_eclasses_"] = {}
            yield cpv, d

    def commit(self):
        self.db.commit()

    def get_matches(self, match_dict):
        query_list = []
        for k, v in match_dict.items():
            if k not in self._known_keys:
                raise cache_errors.InvalidRestriction(
                    k, v, "key isn't known to this cache instance"
                )
            v = v.replace("%", "\\%")
            v = v.replace(".*", "%")
            query_list.append(
                f"(key={self._sfilter(k)} AND value LIKE {self._sfilter(v)})"
            )

        if len(query_list):
            query = " AND " + " AND ".join(query_list)
        else:
            query = ""

        print(
            "query = SELECT cpv from package_cache natural join values_cache WHERE label=%s %s"
            % (self.label, query)
        )
        try:
            self.con.execute(
                "SELECT cpv from package_cache natural join values_cache WHERE label=%s %s"
                % (self.label, query)
            )
        except self._BaseError as e:
            raise cache_errors.GeneralCacheCorruption(e)

        return [row[0] for row in self.con.fetchall()]

    items = iteritems
    keys = __iter__



# test_NewsItem.py -- Portage Unit Testing Functionality
# Copyright 2007-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.news import NewsItem, NewsManager
from portage.dbapi.virtual import fakedbapi

from dataclasses import dataclass
from string import Template
from typing import Optional
from unittest.mock import MagicMock, mock_open, patch

import textwrap

# The specification for news items is GLEP 42 ("Critical News Reporting"):
# https://www.gentoo.org/glep/glep-0042.html


@dataclass
class FakeNewsItem(NewsItem):
    title: str
    author: str
    content_type: str
    posted: str
    revision: int
    news_item_format: str
    content: str
    display_if_installed: Optional[list[str]] = None
    display_if_profile: Optional[list[str]] = None
    display_if_keyword: Optional[list[str]] = None

    item_template_header = Template(
        textwrap.dedent(
            """
        Title: ${title}
        Author: ${author}
        Content-Type: ${content_type}
        Posted: ${posted}
        Revision: ${revision}
        News-Item-Format: ${news_item_format}
        """
        )
    )

    def __post_init__(self):
        super().__init__(path="mocked_news", name=self.title)

    def isValid(self):
        with patch("builtins.open", mock_open(read_data=str(self))):
            return super().isValid()

    # TODO: Migrate __str__ to NewsItem? NewsItem doesn't actually parse
    # all fields right now though.
    def __str__(self) -> str:
        item = self.item_template_header.substitute(
            title=self.title,
            author=self.author,
            content_type=self.content_type,
            posted=self.posted,
            revision=self.revision,
            news_item_format=self.news_item_format,
        )

        for package in self.display_if_installed:
            item += f"Display-If-Installed: {package}\n"

        for profile in self.display_if_profile:
            item += f"Display-If-Profile: {profile}\n"

        for keyword in self.display_if_keyword:
            item += f"Display-If-Keyword: {keyword}\n"

        item += f"\n{self.content}"

        return item


class NewsItemTestCase(TestCase):
    # Default values for testing
    placeholders = {
        "title": "YourSQL Upgrades from 4.0 to 4.1",
        "author": "Ciaran McCreesh <ciaranm@gentoo.org>",
        "content_type": "Content-Type: text/plain",
        "posted": "01-Nov-2005",
        "revision": 1,
        "news_item_format": "1.0",
        "display_if_installed": [],
        "display_if_profile": [],
        "display_if_keyword": [],
        "content": textwrap.dedent(
            """
    YourSQL databases created using YourSQL version 4.0 are incompatible
    with YourSQL version 4.1 or later. There is no reliable way to
    automate the database format conversion, so action from the system
    administrator is required before an upgrade can take place.

    Please see the Gentoo YourSQL Upgrade Guide for instructions:

        https://gentoo.org/doc/en/yoursql-upgrading.xml

    Also see the official YourSQL documentation:

        https://dev.example.com/doc/refman/4.1/en/upgrading-from-4-0.html

    After upgrading, you should also recompile any packages which link
    against YourSQL:

        revdep-rebuild --library=libyoursqlclient.so.12

    The revdep-rebuild tool is provided by app-portage/gentoolkit.
    """
        ),
    }

    def setUp(self) -> None:
        self.profile_base = "/var/db/repos/gentoo/profiles/default-linux"
        self.profile = f"{self.profile_base}/x86/2007.0/"
        self.keywords = "x86"
        # Consumers only use ARCH, so avoid portage.settings by using a dict
        self.settings = {"ARCH": "x86"}
        # Use fake/test dbapi to avoid slow tests
        self.vardb = fakedbapi(self.settings)

    def _createNewsItem(self, *kwargs) -> FakeNewsItem:
        # Use our placeholders unless overridden
        news_args = self.placeholders.copy()
        # Substitute in what we're given to allow for easily passing
        # just custom values.
        news_args.update(*kwargs)

        return FakeNewsItem(**news_args)

    def _checkAndCreateNewsItem(
        self, news_args: dict, relevant: bool = True, reason: str = ""
    ) -> FakeNewsItem:
        return self._checkNewsItem(self._createNewsItem(news_args), relevant, reason)

    def _checkNewsItem(self, item: NewsItem, relevant: bool = True, reason: str = ""):
        self.assertTrue(item.isValid())

        if relevant:
            self.assertTrue(
                item.isRelevant(self.vardb, self.settings, self.profile),
                msg=f"Expected {item} to be relevant, but it was not!",
            )
        else:
            self.assertFalse(
                item.isRelevant(self.vardb, self.settings, self.profile),
                msg=f"Expected {item} to be irrelevant, but it was relevant!",
            )

    def testNewsManager(self):
        vardb = MagicMock()
        portdb = MagicMock()
        portdb.repositories.mainRepoLocation = MagicMock(return_value="/tmp/repo")
        portdb.settings.profile_path = "/tmp/repo/profiles/arch/amd64"

        news_manager = NewsManager(portdb, vardb, portdb.portdir, portdb.portdir)
        self.assertEqual(news_manager._profile_path, "arch/amd64")
        self.assertNotEqual(news_manager._profile_path, "tmp/repo/profiles/arch/amd64")

    def testBasicNewsItem(self):
        # Simple test with no filter fields (Display-If-*)
        item = self._createNewsItem()
        self.assertTrue(item.isValid())
        self.assertTrue(item.isRelevant(self.vardb, self.settings, self.profile))

        # Does an invalid item fail? ("a" is not a valid package name)
        item = self._createNewsItem({"display_if_installed": "a"})
        self.assertFalse(item.isValid())

    def testDisplayIfProfile(self):
        # We repeat all of these with the full profile path (including repo)
        # and a relative path, as we've had issues there before.
        # Note that we can't use _checkNewsItem() here as we override the
        # profile value passed to isRelevant.
        for profile_prefix in ("", self.profile_base):
            # First, just check the simple case of one profile matching ours.
            item = self._createNewsItem(
                {"display_if_profile": [profile_prefix + self.profile]}
            )
            self.assertTrue(item.isValid())
            self.assertTrue(
                item.isRelevant(
                    self.vardb, self.settings, profile_prefix + self.profile
                ),
                msg=f"Expected {item} to be relevant, but it was not!",
            )

            # Test the negative case: what if the only profile listed
            # does *not* match ours?
            item = self._createNewsItem(
                {"display_if_profile": [profile_prefix + "profiles/i-do-not-exist"]}
            )
            self.assertTrue(item.isValid())
            self.assertFalse(
                item.isRelevant(
                    self.vardb, self.settings, profile_prefix + self.profile
                ),
                msg=f"Expected {item} to be irrelevant, but it was relevant!",
            )

            # What if several profiles are listed and we match one of them?
            item = self._createNewsItem(
                {
                    "display_if_profile": [
                        profile_prefix + self.profile,
                        profile_prefix + f"{self.profile_base}/amd64/2023.0",
                    ]
                }
            )
            self.assertTrue(item.isValid())
            self.assertTrue(
                item.isRelevant(
                    self.vardb, self.settings, profile_prefix + self.profile
                ),
                msg=f"Expected {item} to be relevant, but it was not!",
            )

            # What if several profiles are listed and we match none of them?
            item = self._createNewsItem(
                {
                    "display_if_profile": [
                        profile_prefix + f"{self.profile_base}/x86/2023.0",
                        profile_prefix + f"{self.profile_base}/amd64/2023.0",
                    ]
                }
            )
            self.assertTrue(item.isValid())
            self.assertFalse(
                item.isRelevant(
                    self.vardb, self.settings, profile_prefix + self.profile
                ),
                msg=f"Expected {item} to be irrelevant, but it was relevant!",
            )

    def testDisplayIfInstalled(self):
        self.vardb.cpv_inject("sys-apps/portage-2.0", {"SLOT": "0"})

        self._checkAndCreateNewsItem({"display_if_installed": ["sys-apps/portage"]})

        # Test the negative case: a single Display-If-Installed listing
        # a package we don't have.
        self._checkAndCreateNewsItem(
            {"display_if_installed": ["sys-apps/i-do-not-exist"]}, False
        )

        # What about several packages and we have none of them installed?
        self._checkAndCreateNewsItem(
            {
                "display_if_installed": [
                    "dev-util/pkgcheck",
                    "dev-util/pkgdev",
                    "sys-apps/pkgcore",
                ]
            },
            False,
        )

        # What about several packages and we have one of them installed?
        self.vardb.cpv_inject("net-misc/openssh-9.2_p1", {"SLOT": "0"})
        self._checkAndCreateNewsItem(
            {
                "display_if_installed": [
                    "net-misc/openssh",
                    "net-misc/dropbear",
                ]
            }
        )

        # What about several packages and we have all of them installed?
        # Note: we already have openssh added from the above test
        self.vardb.cpv_inject("net-misc/dropbear-2022.83", {"SLOT": "0"})
        self._checkAndCreateNewsItem(
            {
                "display_if_installed": [
                    "net-misc/openssh",
                    "net-misc/dropbear",
                ]
            }
        )

        # What if we have a newer version of the listed package which
        # shouldn't match the constraint?
        self._checkAndCreateNewsItem(
            {
                "display_if_installed": [
                    "<net-misc/openssh-9.2_p1",
                ]
            },
            False,
        )

        # What if we have a newer version of the listed package which
        # should match the constraint?
        self._checkAndCreateNewsItem(
            {
                "display_if_installed": [
                    ">=net-misc/openssh-9.2_p1",
                ]
            }
        )

        # What if the item lists multiple packages and we have one of
        # them installed, but not all?
        # (Note that openssh is already "installed" by this point because
        # of a previous test.)
        self._checkAndCreateNewsItem(
            {
                "display_if_installed": [
                    ">=net-misc/openssh-9.2_p1",
                    "<net-misc/openssh-9.2_p1",
                ]
            }
        )

    def testDisplayIfKeyword(self):
        self._checkAndCreateNewsItem({"display_if_keyword": [self.keywords]})

        # Test the negative case: a keyword we don't have set.
        self._checkAndCreateNewsItem({"display_if_keyword": ["fake-keyword"]}, False)

        # What if several keywords are listed and we match one of them?
        self._checkAndCreateNewsItem(
            {"display_if_keyword": [self.keywords, "amd64", "~hppa"]}
        )

        # What if several keywords are listed and we match none of them?
        self._checkAndCreateNewsItem({"display_if_keyword": ["amd64", "~hppa"]}, False)

        # What if the ~keyword (testing) keyword is listed but we're keyword (stable)?
        self._checkAndCreateNewsItem(
            {
                "display_if_keyword": [
                    f"~{self.keywords}",
                ]
            },
            False,
        )

        # What if the stable keyword is listed but we're ~keyword (testing)?
        self._checkAndCreateNewsItem(
            {
                "display_if_keyword": [
                    f"{self.keywords}",
                ]
            }
        )

    def testMultipleRestrictions(self):
        # GLEP 42 specifies an algorithm for how combining restrictions
        # should work. See https://www.gentoo.org/glep/glep-0042.html#news-item-headers.
        # Different types of Display-If-* are ANDed, not ORed.

        # What if there's a Display-If-Keyword that matches and a
        # Display-If-Installed which does too?
        self.vardb.cpv_inject("sys-apps/portage-2.0", {"SLOT": "0"})
        self._checkAndCreateNewsItem(
            {
                "display_if_keyword": [self.keywords],
                "display_if_installed": ["sys-apps/portage"],
            }
        )

        # What if there's a Display-If-Keyword that matches and a
        # Display-If-Installed which doesn't?
        self._checkAndCreateNewsItem(
            {
                "display_if_keyword": [self.keywords],
                "display_if_installed": ["sys-apps/i-do-not-exist"],
            },
            False,
        )

        # What if there's a Display-If-{Installed,Keyword,Profile} and
        # they all match?
        # (Note that sys-apps/portage is already "installed" by this point
        # because of the above test.)
        self._checkAndCreateNewsItem(
            {
                "display_if_keyword": [self.keywords],
                "display_if_installed": ["sys-apps/portage"],
                "display_if_profile": [self.profile],
            }
        )

        # What if there's a Display-If-{Installed,Keyword,Profile} and
        # none of them match?
        # (Note that sys-apps/portage is already "installed" by this point
        # because of the above test.)
        self._checkAndCreateNewsItem(
            {
                "display_if_keyword": ["i-do-not-exist"],
                "display_if_installed": ["sys-apps/i-do-not-exist"],
                "display_if_profile": [self.profile_base + "/i-do-not-exist"],
            },
            False,
        )



# Copyright 2022 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class VariableSetTestCase(TestCase):
    def testVariableSetEmerge(self):
        ebuilds = {
            "dev-go/go-pkg-1": {"BDEPEND": "dev-lang/go"},
        }
        installed = ebuilds
        playground = ResolverPlayground(ebuilds=ebuilds, installed=installed)

        test_cases = (
            ResolverPlaygroundTestCase(
                ["@golang-rebuild"],
                mergelist=["dev-go/go-pkg-1"],
                success=True,
            ),
        )

        try:
            for test_case in test_cases:
                # Create an artificial VariableSet to test against
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2014-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

"""
Base class for performing sync operations.
This class contains common initialization code and functions.
"""

import functools
import logging
import os

import portage
from portage.repository.storage.interface import RepoStorageException
from portage.util import writemsg_level
from portage.util._eventloop.global_event_loop import global_event_loop
from portage.util.backoff import RandomExponentialBackoff
from portage.util.futures._sync_decorator import _sync_methods
from portage.util.futures.retry import retry
from portage.util.futures.executor.fork import ForkExecutor
from . import _SUBMODULE_PATH_MAP

try:
    import gemato.openpgp
except ImportError:
    gemato = None


class SyncBase:
    """Base Sync class for subclassing"""

    short_desc = "Perform sync operations on repositories"

    @staticmethod
    def name():
        return "BlankSync"

    def can_progressbar(self, func):
        return False

    def __init__(self, bin_command, bin_pkg):
        self.options = None
        self.settings = None
        self.logger = None
        self.repo = None
        self.xterm_titles = None
        self.spawn_kwargs = None
        self._repo_storage = None
        self._download_dir = None
        self.bin_command = None
        self._bin_command = bin_command
        self.bin_pkg = bin_pkg
        if bin_command:
            self.bin_command = portage.process.find_binary(bin_command)

    @property
    def has_bin(self):
        """Checks for existance of the external binary, and also
        checks for storage driver configuration problems.

        MUST only be called after _kwargs() has set the logger
        """
        if self.bin_command is None:
            msg = [
                f"Command not found: {self._bin_command}",
                f'Type "emerge {self.bin_pkg}" to enable {self._bin_command} support.',
            ]
            for l in msg:
                writemsg_level(f"!!! {l}\n", level=logging.ERROR, noiselevel=-1)
            return False

        try:
            self.repo_storage
        except RepoStorageException as e:
            writemsg_level(f"!!! {e}\n", level=logging.ERROR, noiselevel=-1)
            return False

        return True

    def _kwargs(self, kwargs):
        """Sets internal variables from kwargs"""
        self.options = kwargs.get("options", {})
        self.settings = self.options.get("settings", None)
        self.logger = self.options.get("logger", None)
        self.repo = self.options.get("repo", None)
        self.xterm_titles = self.options.get("xterm_titles", False)
        self.spawn_kwargs = self.options.get("spawn_kwargs", None)

    def _select_storage_module(self):
        """
        Select an appropriate implementation of RepoStorageInterface, based
        on repos.conf settings.

        @rtype: str
        @return: name of the selected repo storage constructor
        """
        if self.repo.sync_rcu:
            mod_name = "portage.repository.storage.hardlink_rcu.HardlinkRcuRepoStorage"
        elif self.repo.sync_allow_hardlinks:
            mod_name = "portage.repository.storage.hardlink_quarantine.HardlinkQuarantineRepoStorage"
        else:
            mod_name = "portage.repository.storage.inplace.InplaceRepoStorage"
        return mod_name

    @property
    def repo_storage(self):
        """
        Get the repo storage driver instance. Raise RepoStorageException
        if there is a configuration problem
        """
        if self._repo_storage is None:
            storage_cls = portage.load_mod(self._select_storage_module())
            self._repo_storage = _sync_methods(
                storage_cls(self.repo, self.spawn_kwargs), loop=global_event_loop()
            )
        return self._repo_storage

    @property
    def download_dir(self):
        """
        Get the path of the download directory, where the repository
        update is staged. The directory is initialized lazily, since
        the repository might already be at the latest revision, and
        there may be some cost associated with the directory
        initialization.
        """
        if self._download_dir is None:
            self._download_dir = self.repo_storage.init_update()
        return self._download_dir

    def exists(self, **kwargs):
        """Tests whether the repo actually exists"""
        if kwargs:
            self._kwargs(kwargs)
        elif not self.repo:
            return False
        if not os.path.exists(self.repo.location):
            return False
        return True

    def sync(self, **kwargs):
        """Sync the repository"""
        raise NotImplementedError

    def post_sync(self, portdb, location, emerge_config):
        """repo.sync_type == "Blank":
        # NOTE: Do this after reloading the config, in case
        # it did not exist prior to sync, so that the config
        # and portdb properly account for its existence.
        """
        pass

    def _get_submodule_paths(self):
        paths = []
        emerge_config = self.options.get("emerge_config")
        if emerge_config is not None:
            for name in emerge_config.opts.get("--sync-submodule", []):
                paths.extend(_SUBMODULE_PATH_MAP[name])
        return tuple(paths)

    def retrieve_head(self, **kwargs):
        """Get information about the head commit"""
        raise NotImplementedError

    def _key_refresh_retry_decorator(self):
        """
        Return a retry decorator, or None if retry is disabled.

        If retry fails, the function reraises the exception raised
        by the decorated function. If retry times out and no exception
        is available to reraise, the function raises TimeoutError.
        """
        errors = []

        if self.repo.sync_openpgp_key_refresh_retry_count is None:
            return None
        try:
            retry_count = int(self.repo.sync_openpgp_key_refresh_retry_count)
        except Exception as e:
            errors.append(f"sync-openpgp-key-refresh-retry-count: {e}")
        else:
            if retry_count <= 0:
                return None

        if self.repo.sync_openpgp_key_refresh_retry_overall_timeout is None:
            retry_overall_timeout = None
        else:
            try:
                retry_overall_timeout = float(
                    self.repo.sync_openpgp_key_refresh_retry_overall_timeout
                )
            except Exception as e:
                errors.append(f"sync-openpgp-key-refresh-retry-overall-timeout: {e}")
            else:
                if retry_overall_timeout < 0:
                    errors.append(
                        "sync-openpgp-key-refresh-retry-overall-timeout: "
                        "value must be greater than or equal to zero: {}".format(
                            retry_overall_timeout
                        )
                    )
                elif retry_overall_timeout == 0:
                    retry_overall_timeout = None

        if self.repo.sync_openpgp_key_refresh_retry_delay_mult is None:
            retry_delay_mult = None
        else:
            try:
                retry_delay_mult = float(
                    self.repo.sync_openpgp_key_refresh_retry_delay_mult
                )
            except Exception as e:
                errors.append(f"sync-openpgp-key-refresh-retry-delay-mult: {e}")
            else:
                if retry_delay_mult <= 0:
                    errors.append(
                        "sync-openpgp-key-refresh-retry-mult: "
                        "value must be greater than zero: {}".format(retry_delay_mult)
                    )

        if self.repo.sync_openpgp_key_refresh_retry_delay_exp_base is None:
            retry_delay_exp_base = None
        else:
            try:
                retry_delay_exp_base = float(
                    self.repo.sync_openpgp_key_refresh_retry_delay_exp_base
                )
            except Exception as e:
                errors.append(f"sync-openpgp-key-refresh-retry-delay-exp: {e}")
            else:
                if retry_delay_exp_base <= 0:
                    errors.append(
                        "sync-openpgp-key-refresh-retry-delay-exp: "
                        "value must be greater than zero: {}".format(retry_delay_mult)
                    )

        if errors:
            lines = []
            lines.append("")
            lines.append("!!! Retry disabled for openpgp key refresh:")
            lines.append("")
            for msg in errors:
                lines.append(f"    {msg}")
            lines.append("")

            for line in lines:
                writemsg_level(f"{line}\n", level=logging.ERROR, noiselevel=-1)

            return None

        return retry(
            reraise=True,
            try_max=retry_count,
            overall_timeout=(
                retry_overall_timeout if retry_overall_timeout > 0 else None
            ),
            delay_func=RandomExponentialBackoff(
                multiplier=(1 if retry_delay_mult is None else retry_delay_mult),
                base=(2 if retry_delay_exp_base is None else retry_delay_exp_base),
            ),
        )

    def _refresh_keys(self, openpgp_env):
        """
        Refresh keys stored in openpgp_env. Raises gemato.exceptions.GematoException
        or asyncio.TimeoutError on failure.

        @param openpgp_env: openpgp environment
        @type openpgp_env: gemato.openpgp.OpenPGPEnvironment
        """
        out = portage.output.EOutput(
            quiet=("--quiet" in self.options["emerge_config"].opts)
        )

        if not self.repo.sync_openpgp_key_refresh:
            out.ewarn(
                "Key refresh is disabled via a repos.conf sync-openpgp-key-refresh"
            )
            out.ewarn(
                "setting, and this is a security vulnerability because it prevents"
            )
            out.ewarn("detection of revoked keys!")
            return

        out.ebegin("Refreshing keys via WKD")
        if openpgp_env.refresh_keys_wkd():
            out.eend(0)
            return
        out.eend(1)

        out.ebegin(
            "Refreshing keys from keyserver{}".format(
                ""
                if self.repo.sync_openpgp_keyserver is None
                else " " + self.repo.sync_openpgp_keyserver
            )
        )
        retry_decorator = self._key_refresh_retry_decorator()
        if retry_decorator is None:
            openpgp_env.refresh_keys_keyserver(
                keyserver=self.repo.sync_openpgp_keyserver
            )
        else:

            def noisy_refresh_keys():
                """
                Since retry does not help for some types of
                errors, display errors as soon as they occur.
                """
                try:
                    openpgp_env.refresh_keys_keyserver(
                        keyserver=self.repo.sync_openpgp_keyserver
                    )
                except Exception as e:
                    writemsg_level(f"{e}\n", level=logging.ERROR, noiselevel=-1)
                    raise  # retry

            # The ThreadPoolExecutor that asyncio uses by default
            # does not support cancellation of tasks, therefore
            # use ForkExecutor for task cancellation support, in
            # order to enforce timeouts.
            loop = global_event_loop()
            with ForkExecutor(loop=loop) as executor:
                func_coroutine = functools.partial(
                    loop.run_in_executor, executor, noisy_refresh_keys
                )
                decorated_func = retry_decorator(func_coroutine, loop=loop)
                loop.run_until_complete(decorated_func())
        out.eend(0)

    def _get_openpgp_env(self, openpgp_key_path=None, debug=False):
        if gemato is not None:
            # Override global proxy setting with one provided in emerge configuration
            if "http_proxy" in self.spawn_kwargs["env"]:
                proxy = self.spawn_kwargs["env"]["http_proxy"]
            elif "https_proxy" in self.spawn_kwargs["env"]:
                proxy = self.spawn_kwargs["env"]["https_proxy"]
            else:
                proxy = None

            if openpgp_key_path:
                openpgp_env = gemato.openpgp.OpenPGPEnvironment(
                    proxy=proxy, debug=debug
                )
            else:
                openpgp_env = gemato.openpgp.OpenPGPSystemEnvironment(
                    proxy=proxy, debug=debug
                )

            return openpgp_env


class NewBase(SyncBase):
    """Subclasses Syncbase adding a new() and runs it
    instead of update() if the repository does not exist()"""

    def __init__(self, bin_command, bin_pkg):
        SyncBase.__init__(self, bin_command, bin_pkg)

    def sync(self, **kwargs):
        """Sync the repository"""
        if kwargs:
            self._kwargs(kwargs)

        if not self.has_bin:
            return (1, False)

        if not self.exists():
            return self.new()
        return self.update()

    def new(self, **kwargs):
        """Do the initial download and install of the repository"""
        raise NotImplementedError

    def update(self):
        """Update existing repository"""
        raise NotImplementedError



# Copyright 2010-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class AutounmaskTestCase(TestCase):
    def testAutounmask(self):
        ebuilds = {
            # ebuilds to test use changes
            "dev-libs/A-1": {"SLOT": 1, "DEPEND": "dev-libs/B[foo]", "EAPI": 2},
            "dev-libs/A-2": {"SLOT": 2, "DEPEND": "dev-libs/B[bar]", "EAPI": 2},
            "dev-libs/B-1": {
                "DEPEND": "foo? ( dev-libs/C ) bar? ( dev-libs/D )",
                "IUSE": "foo bar",
            },
            "dev-libs/C-1": {},
            "dev-libs/D-1": {},
            # ebuilds to test if we allow changing of masked or forced flags
            "dev-libs/E-1": {"SLOT": 1, "DEPEND": "dev-libs/F[masked-flag]", "EAPI": 2},
            "dev-libs/E-2": {
                "SLOT": 2,
                "DEPEND": "dev-libs/G[-forced-flag]",
                "EAPI": 2,
            },
            "dev-libs/F-1": {"IUSE": "masked-flag"},
            "dev-libs/G-1": {"IUSE": "forced-flag"},
            # ebuilds to test keyword changes
            "app-misc/Z-1": {"KEYWORDS": "~x86", "DEPEND": "app-misc/Y"},
            "app-misc/Y-1": {"KEYWORDS": "~x86"},
            "app-misc/W-1": {},
            "app-misc/W-2": {"KEYWORDS": "~x86"},
            "app-misc/V-1": {"KEYWORDS": "~x86", "DEPEND": ">=app-misc/W-2"},
            # ebuilds to test mask and keyword changes
            "app-text/A-1": {},
            "app-text/B-1": {"KEYWORDS": "~x86"},
            "app-text/C-1": {"KEYWORDS": ""},
            "app-text/D-1": {"KEYWORDS": "~x86"},
            "app-text/D-2": {"KEYWORDS": ""},
            # ebuilds for mixed test for || dep handling
            "sci-libs/K-1": {
                "DEPEND": " || ( sci-libs/L[bar] || ( sci-libs/M sci-libs/P ) )",
                "EAPI": 2,
            },
            "sci-libs/K-2": {
                "DEPEND": " || ( sci-libs/L[bar] || ( sci-libs/P sci-libs/M ) )",
                "EAPI": 2,
            },
            "sci-libs/K-3": {
                "DEPEND": " || ( sci-libs/M || ( sci-libs/L[bar] sci-libs/P ) )",
                "EAPI": 2,
            },
            "sci-libs/K-4": {
                "DEPEND": " || ( sci-libs/M || ( sci-libs/P sci-libs/L[bar] ) )",
                "EAPI": 2,
            },
            "sci-libs/K-5": {
                "DEPEND": " || ( sci-libs/P || ( sci-libs/L[bar] sci-libs/M ) )",
                "EAPI": 2,
            },
            "sci-libs/K-6": {
                "DEPEND": " || ( sci-libs/P || ( sci-libs/M sci-libs/L[bar] ) )",
                "EAPI": 2,
            },
            "sci-libs/K-7": {"DEPEND": " || ( sci-libs/M sci-libs/L[bar] )", "EAPI": 2},
            "sci-libs/K-8": {"DEPEND": " || ( sci-libs/L[bar] sci-libs/M )", "EAPI": 2},
            "sci-libs/L-1": {"IUSE": "bar"},
            "sci-libs/M-1": {"KEYWORDS": "~x86"},
            "sci-libs/P-1": {},
            # ebuilds to test these nice "required by cat/pkg[foo]" messages
            "dev-util/Q-1": {
                "DEPEND": "foo? ( dev-util/R[bar] )",
                "IUSE": "+foo",
                "EAPI": 2,
            },
            "dev-util/Q-2": {
                "RDEPEND": "!foo? ( dev-util/R[bar] )",
                "IUSE": "foo",
                "EAPI": 2,
            },
            "dev-util/R-1": {"IUSE": "bar"},
            # ebuilds to test interaction with REQUIRED_USE
            "app-portage/A-1": {"DEPEND": "app-portage/B[foo]", "EAPI": 2},
            "app-portage/A-2": {
                "DEPEND": "app-portage/B[foo=]",
                "IUSE": "+foo",
                "REQUIRED_USE": "foo",
                "EAPI": "4",
            },
            "app-portage/B-1": {
                "IUSE": "foo +bar",
                "REQUIRED_USE": "^^ ( foo bar )",
                "EAPI": "4",
            },
            "app-portage/C-1": {
                "IUSE": "+foo +bar",
                "REQUIRED_USE": "^^ ( foo bar )",
                "EAPI": "4",
            },
            "sci-mathematics/octave-4.2.2": {
                "EAPI": 6,
                "RDEPEND": ">=x11-libs/qscintilla-2.9.3-r2:=[qt5(+)]",
            },
            "x11-libs/qscintilla-2.9.4": {
                "EAPI": 6,
                "IUSE": "+qt4 qt5",
                "REQUIRED_USE": "^^ ( qt4 qt5 )",
            },
            "x11-libs/qscintilla-2.10": {
                "EAPI": 6,
                "KEYWORDS": "~x86",
                "IUSE": "qt4 +qt5",
            },
        }

        test_cases = (
            # Test USE changes.
            # The simple case.
            ResolverPlaygroundTestCase(
                ["dev-libs/A:1"], options={"--autounmask": "n"}, success=False
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/A:1"],
                options={"--autounmask": True},
                success=False,
                mergelist=["dev-libs/C-1", "dev-libs/B-1", "dev-libs/A-1"],
                use_changes={"dev-libs/B-1": {"foo": True}},
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/A:1"],
                options={"--autounmask-use": "y"},
                success=False,
                mergelist=["dev-libs/C-1", "dev-libs/B-1", "dev-libs/A-1"],
                use_changes={"dev-libs/B-1": {"foo": True}},
            ),
            # Test default --autounmask-use
            ResolverPlaygroundTestCase(
                ["dev-libs/A:1"],
                success=False,
                mergelist=["dev-libs/C-1", "dev-libs/B-1", "dev-libs/A-1"],
                use_changes={"dev-libs/B-1": {"foo": True}},
            ),
            # Explicitly disable --autounmask-use
            ResolverPlaygroundTestCase(
                ["dev-libs/A:1"], success=False, options={"--autounmask-use": "n"}
            ),
            # Make sure we restart if needed.
            ResolverPlaygroundTestCase(
                ["dev-libs/A:1", "dev-libs/B"],
                options={"--autounmask": True, "--autounmask-backtrack": "y"},
                all_permutations=True,
                success=False,
                mergelist=["dev-libs/C-1", "dev-libs/B-1", "dev-libs/A-1"],
                use_changes={"dev-libs/B-1": {"foo": True}},
            ),
            # With --autounmask-backtrack=y:
            # [ebuild  N     ] dev-libs/C-1
            # [ebuild  N     ] dev-libs/B-1  USE="foo -bar"
            # [ebuild  N     ] dev-libs/A-1
            #
            # The following USE changes are necessary to proceed:
            # (see "package.use" in the portage(5) man page for more details)
            # required by dev-libs/A-1::test_repo
            # required by dev-libs/A:1 (argument)
            # >=dev-libs/B-1 foo
            # Without --autounmask-backtrack=y:
            # [ebuild  N     ] dev-libs/B-1  USE="foo -bar"
            # [ebuild  N     ] dev-libs/A-1
            #
            # The following USE changes are necessary to proceed:
            # (see "package.use" in the portage(5) man page for more details)
            # required by dev-libs/A-1::test_repo
            # required by dev-libs/A:1 (argument)
            # >=dev-libs/B-1 foo
            ResolverPlaygroundTestCase(
                ["dev-libs/A:1", "dev-libs/A:2", "dev-libs/B"],
                options={"--autounmask": True, "--autounmask-backtrack": "y"},
                all_permutations=True,
                success=False,
                mergelist=[
                    "dev-libs/D-1",
                    "dev-libs/C-1",
                    "dev-libs/B-1",
                    "dev-libs/A-1",
                    "dev-libs/A-2",
                ],
                ignore_mergelist_order=True,
                use_changes={"dev-libs/B-1": {"foo": True, "bar": True}},
            ),
            # With --autounmask-backtrack=y:
            # [ebuild  N     ] dev-libs/C-1
            # [ebuild  N     ] dev-libs/D-1
            # [ebuild  N     ] dev-libs/B-1  USE="bar foo"
            # [ebuild  N     ] dev-libs/A-2
            # [ebuild  N     ] dev-libs/A-1
            #
            # The following USE changes are necessary to proceed:
            # (see "package.use" in the portage(5) man page for more details)
            # required by dev-libs/A-2::test_repo
            # required by dev-libs/A:2 (argument)
            # >=dev-libs/B-1 bar foo
            # Without --autounmask-backtrack=y:
            # [ebuild  N     ] dev-libs/B-1  USE="bar foo"
            # [ebuild  N     ] dev-libs/A-1
            # [ebuild  N     ] dev-libs/A-2
            #
            # The following USE changes are necessary to proceed:
            # (see "package.use" in the portage(5) man page for more details)
            # required by dev-libs/A-1::test_repo
            # required by dev-libs/A:1 (argument)
            # >=dev-libs/B-1 foo bar
            # NOTE: The --autounmask-backtrack=n behavior is acceptable, but
            # it would be nicer if it added the dev-libs/C-1 and dev-libs/D-1
            # deps to the depgraph without backtracking. It could add two
            # instances of dev-libs/B-1 to the graph with different USE flags,
            # and then use _solve_non_slot_operator_slot_conflicts to eliminate
            # the redundant instance.
            # Test keywording.
            # The simple case.
            ResolverPlaygroundTestCase(
                ["app-misc/Z"], options={"--autounmask": "n"}, success=False
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/Z"],
                options={"--autounmask": True},
                success=False,
                mergelist=["app-misc/Y-1", "app-misc/Z-1"],
                unstable_keywords=["app-misc/Y-1", "app-misc/Z-1"],
            ),
            # Make sure that the backtracking for slot conflicts handles our mess.
            ResolverPlaygroundTestCase(
                ["=app-misc/V-1", "app-misc/W"],
                options={"--autounmask": True},
                all_permutations=True,
                success=False,
                mergelist=["app-misc/W-2", "app-misc/V-1"],
                unstable_keywords=["app-misc/W-2", "app-misc/V-1"],
            ),
            # Mixed testing
            # Make sure we don't change use for something in a || dep if there is another choice
            # that needs no change.
            ResolverPlaygroundTestCase(
                ["=sci-libs/K-1"],
                options={"--autounmask": True},
                success=True,
                mergelist=["sci-libs/P-1", "sci-libs/K-1"],
            ),
            ResolverPlaygroundTestCase(
                ["=sci-libs/K-2"],
                options={"--autounmask": True},
                success=True,
                mergelist=["sci-libs/P-1", "sci-libs/K-2"],
            ),
            ResolverPlaygroundTestCase(
                ["=sci-libs/K-3"],
                options={"--autounmask": True},
                success=True,
                mergelist=["sci-libs/P-1", "sci-libs/K-3"],
            ),
            ResolverPlaygroundTestCase(
                ["=sci-libs/K-4"],
                options={"--autounmask": True},
                success=True,
                mergelist=["sci-libs/P-1", "sci-libs/K-4"],
            ),
            ResolverPlaygroundTestCase(
                ["=sci-libs/K-5"],
                options={"--autounmask": True},
                success=True,
                mergelist=["sci-libs/P-1", "sci-libs/K-5"],
            ),
            ResolverPlaygroundTestCase(
                ["=sci-libs/K-6"],
                options={"--autounmask": True},
                success=True,
                mergelist=["sci-libs/P-1", "sci-libs/K-6"],
            ),
            # Make sure we prefer use changes over keyword changes.
            ResolverPlaygroundTestCase(
                ["=sci-libs/K-7"],
                options={"--autounmask": True},
                success=False,
                mergelist=["sci-libs/L-1", "sci-libs/K-7"],
                use_changes={"sci-libs/L-1": {"bar": True}},
            ),
            ResolverPlaygroundTestCase(
                ["=sci-libs/K-8"],
                options={"--autounmask": True},
                success=False,
                mergelist=["sci-libs/L-1", "sci-libs/K-8"],
                use_changes={"sci-libs/L-1": {"bar": True}},
            ),
            # Test these nice "required by cat/pkg[foo]" messages.
            ResolverPlaygroundTestCase(
                ["=dev-util/Q-1"],
                options={"--autounmask": True},
                success=False,
                mergelist=["dev-util/R-1", "dev-util/Q-1"],
                use_changes={"dev-util/R-1": {"bar": True}},
            ),
            ResolverPlaygroundTestCase(
                ["=dev-util/Q-2"],
                options={"--autounmask": True},
                success=False,
                mergelist=["dev-util/R-1", "dev-util/Q-2"],
                use_changes={"dev-util/R-1": {"bar": True}},
            ),
            # Test interaction with REQUIRED_USE.
            # Some of these cases trigger USE change(s) that violate
            # REQUIRED_USE, so the USE changes are shown along with
            # the REQUIRED_USE violation that they would trigger.
            # The following USE changes are necessary to proceed:
            #  (see "package.use" in the portage(5) man page for more details)
            # # required by app-portage/A-1::test_repo
            # # required by =app-portage/A-1 (argument)
            # >=app-portage/B-1 foo
            #
            # !!! The ebuild selected to satisfy "app-portage/B[foo]" has unmet requirements.
            # - app-portage/B-1::test_repo USE="bar (forced-flag) -foo"
            #
            #   The following REQUIRED_USE flag constraints are unsatisfied:
            #     exactly-one-of ( foo bar )
            ResolverPlaygroundTestCase(
                ["=app-portage/A-1"],
                options={"--autounmask": True},
                use_changes={"app-portage/B-1": {"foo": True}},
                success=False,
            ),
            # The following USE changes are necessary to proceed:
            #  (see "package.use" in the portage(5) man page for more details)
            # # required by app-portage/A-2::test_repo
            # # required by =app-portage/A-2 (argument)
            # >=app-portage/B-1 foo
            #
            # !!! The ebuild selected to satisfy "app-portage/B[foo=]" has unmet requirements.
            # - app-portage/B-1::test_repo USE="bar (forced-flag) -foo"
            #
            #   The following REQUIRED_USE flag constraints are unsatisfied:
            #     exactly-one-of ( foo bar )
            ResolverPlaygroundTestCase(
                ["=app-portage/A-2"],
                options={"--autounmask": True},
                use_changes={"app-portage/B-1": {"foo": True}},
                success=False,
            ),
            ResolverPlaygroundTestCase(
                ["=app-portage/C-1"],
                options={"--autounmask": True},
                use_changes=None,
                success=False,
            ),
            # Test bug 622462, where it inappropriately unmasked a newer
            # version rather than report unsatisfied REQUIRED_USE.
            #
            # The following USE changes are necessary to proceed:
            #  (see "package.use" in the portage(5) man page for more details)
            # # required by sci-mathematics/octave-4.2.2::test_repo
            # # required by sci-mathematics/octave (argument)
            # >=x11-libs/qscintilla-2.9.4 qt5
            #
            # !!! The ebuild selected to satisfy ">=x11-libs/qscintilla-2.9.3-r2:=[qt5(+)]" has unmet requirements.
            # - x11-libs/qscintilla-2.9.4::test_repo USE="qt4 -qt5"
            #
            #   The following REQUIRED_USE flag constraints are unsatisfied:
            #     exactly-one-of ( qt4 qt5 )
            #
            # (dependency required by "sci-mathematics/octave-4.2.2::test_repo" [ebuild])
            # (dependency required by "sci-mathematics/octave" [argument])
            ResolverPlaygroundTestCase(
                ["sci-mathematics/octave"],
                options={"--autounmask": True},
                use_changes={"x11-libs/qscintilla-2.9.4": {"qt5": True}},
                success=False,
            ),
            # Make sure we don't change masked/forced flags.
            ResolverPlaygroundTestCase(
                ["dev-libs/E:1"],
                options={"--autounmask": True},
                use_changes=None,
                success=False,
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/E:2"],
                options={"--autounmask": True},
                use_changes=None,
                success=False,
            ),
            # Test mask and keyword changes.
            ResolverPlaygroundTestCase(
                ["app-text/A"],
                options={"--autounmask": True},
                success=False,
                mergelist=["app-text/A-1"],
                needed_p_mask_changes=["app-text/A-1"],
            ),
            ResolverPlaygroundTestCase(
                ["app-text/B"],
                options={"--autounmask": True},
                success=False,
                mergelist=["app-text/B-1"],
                unstable_keywords=["app-text/B-1"],
                needed_p_mask_changes=["app-text/B-1"],
            ),
            ResolverPlaygroundTestCase(
                ["app-text/C"],
                options={"--autounmask": True},
                success=False,
                mergelist=["app-text/C-1"],
                unstable_keywords=["app-text/C-1"],
                needed_p_mask_changes=["app-text/C-1"],
            ),
            # Make sure unstable keyword is preferred over missing keyword
            ResolverPlaygroundTestCase(
                ["app-text/D"],
                options={"--autounmask": True},
                success=False,
                mergelist=["app-text/D-1"],
                unstable_keywords=["app-text/D-1"],
            ),
            # Test missing keyword
            ResolverPlaygroundTestCase(
                ["=app-text/D-2"],
                options={"--autounmask": True},
                success=False,
                mergelist=["app-text/D-2"],
                unstable_keywords=["app-text/D-2"],
            ),
        )

        profile = {
            "use.mask": ("masked-flag",),
            "use.force": ("forced-flag",),
            "package.mask": (
                "app-text/A",
                "app-text/B",
                "app-text/C",
            ),
        }

        playground = ResolverPlayground(ebuilds=ebuilds, profile=profile)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testAutounmaskForLicenses(self):
        ebuilds = {
            "dev-libs/A-1": {"LICENSE": "TEST"},
            "dev-libs/B-1": {"LICENSE": "TEST", "IUSE": "foo", "KEYWORDS": "~x86"},
            "dev-libs/C-1": {"DEPEND": "dev-libs/B[foo]", "EAPI": 2},
            "dev-libs/D-1": {"DEPEND": "dev-libs/E dev-libs/F", "LICENSE": "TEST"},
            "dev-libs/E-1": {"LICENSE": "TEST"},
            "dev-libs/E-2": {"LICENSE": "TEST"},
            "dev-libs/F-1": {"DEPEND": "=dev-libs/E-1", "LICENSE": "TEST"},
            "dev-java/sun-jdk-1.6.0.32": {"LICENSE": "TEST", "KEYWORDS": "~x86"},
            "dev-java/sun-jdk-1.6.0.31": {"LICENSE": "TEST", "KEYWORDS": "x86"},
        }

        test_cases = (
            # --autounmask=n negates default --autounmask-license
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-1"], options={"--autounmask": "n"}, success=False
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-1"],
                options={"--autounmask-license": "y"},
                success=False,
                mergelist=["dev-libs/A-1"],
                license_changes={"dev-libs/A-1": {"TEST"}},
            ),
            # Test that --autounmask enables --autounmask-license
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-1"],
                options={"--autounmask": True},
                success=False,
                mergelist=["dev-libs/A-1"],
                license_changes={"dev-libs/A-1": {"TEST"}},
            ),
            # Test that --autounmask-license is not enabled by default
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-1"],
                success=False,
            ),
            # Test that --autounmask does not override --autounmask-license=n
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-1"],
                options={"--autounmask": True, "--autounmask-license": "n"},
                success=False,
            ),
            # Test that --autounmask=n overrides --autounmask-license=y
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-1"],
                options={"--autounmask": "n", "--autounmask-license": "y"},
                success=False,
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-1"], options={"--autounmask-license": "n"}, success=False
            ),
            # Test license+keyword+use change at once.
            ResolverPlaygroundTestCase(
                ["=dev-libs/C-1"],
                options={"--autounmask": True},
                success=False,
                mergelist=["dev-libs/B-1", "dev-libs/C-1"],
                license_changes={"dev-libs/B-1": {"TEST"}},
                unstable_keywords=["dev-libs/B-1"],
                use_changes={"dev-libs/B-1": {"foo": True}},
            ),
            # Test license with backtracking.
            ResolverPlaygroundTestCase(
                ["=dev-libs/D-1"],
                options={"--autounmask": True},
                success=False,
                mergelist=["dev-libs/E-1", "dev-libs/F-1", "dev-libs/D-1"],
                license_changes={
                    "dev-libs/D-1": {"TEST"},
                    "dev-libs/E-1": {"TEST"},
                    "dev-libs/E-2": {"TEST"},
                    "dev-libs/F-1": {"TEST"},
                },
            ),
            # Test license only for bug #420847
            ResolverPlaygroundTestCase(
                ["dev-java/sun-jdk"],
                options={"--autounmask": True},
                success=False,
                mergelist=["dev-java/sun-jdk-1.6.0.31"],
                license_changes={"dev-java/sun-jdk-1.6.0.31": {"TEST"}},
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testAutounmaskAndSets(self):
        ebuilds = {
            # ebuilds to test use changes
            "dev-libs/A-1": {},
            "dev-libs/A-2": {"KEYWORDS": "~x86"},
            "dev-libs/B-1": {"DEPEND": "dev-libs/A"},
            "dev-libs/C-1": {"DEPEND": ">=dev-libs/A-2"},
            "dev-libs/D-1": {"DEPEND": "dev-libs/A"},
        }

        world_sets = ["@test-set"]
        sets = {
            "test-set": (
                "dev-libs/A",
                "dev-libs/B",
                "dev-libs/C",
                "dev-libs/D",
            ),
        }

        test_cases = (
            # Test USE changes.
            # The simple case.
            ResolverPlaygroundTestCase(
                ["dev-libs/B", "dev-libs/C", "dev-libs/D"],
                all_permutations=True,
                options={"--autounmask": "y"},
                mergelist=[
                    "dev-libs/A-2",
                    "dev-libs/B-1",
                    "dev-libs/C-1",
                    "dev-libs/D-1",
                ],
                ignore_mergelist_order=True,
                unstable_keywords=["dev-libs/A-2"],
                success=False,
            ),
            ResolverPlaygroundTestCase(
                ["@test-set"],
                all_permutations=True,
                options={"--autounmask": "y"},
                mergelist=[
                    "dev-libs/A-2",
                    "dev-libs/B-1",
                    "dev-libs/C-1",
                    "dev-libs/D-1",
                ],
                ignore_mergelist_order=True,
                unstable_keywords=["dev-libs/A-2"],
                success=False,
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                all_permutations=True,
                options={"--autounmask": "y"},
                mergelist=[
                    "dev-libs/A-2",
                    "dev-libs/B-1",
                    "dev-libs/C-1",
                    "dev-libs/D-1",
                ],
                ignore_mergelist_order=True,
                unstable_keywords=["dev-libs/A-2"],
                success=False,
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, world_sets=world_sets, sets=sets
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testAutounmaskKeepMasks(self):
        """
        Ensure that we try to use a masked version with keywords before trying
        masked version with missing keywords (prefer masked regular version
        over -9999 version).
        """
        ebuilds = {
            "app-text/A-1": {},
        }

        test_cases = (
            # Test mask and keyword changes.
            ResolverPlaygroundTestCase(
                ["app-text/A"],
                options={"--autounmask": True, "--autounmask-keep-masks": "y"},
                success=False,
            ),
            ResolverPlaygroundTestCase(
                ["app-text/A"],
                options={"--autounmask": True, "--autounmask-keep-masks": "n"},
                success=False,
                mergelist=["app-text/A-1"],
                needed_p_mask_changes=["app-text/A-1"],
            ),
        )

        profile = {
            "package.mask": ("app-text/A",),
        }

        playground = ResolverPlayground(ebuilds=ebuilds, profile=profile)

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testAutounmask9999(self):
        ebuilds = {
            "dev-libs/A-1": {},
            "dev-libs/A-2": {},
            "dev-libs/A-9999": {"KEYWORDS": ""},
            "dev-libs/B-1": {"DEPEND": ">=dev-libs/A-2"},
            "dev-libs/C-1": {"DEPEND": ">=dev-libs/A-3"},
        }

        profile = {
            "package.mask": (">=dev-libs/A-2",),
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/B"],
                success=False,
                options={"--autounmask": True},
                mergelist=["dev-libs/A-2", "dev-libs/B-1"],
                needed_p_mask_changes={"dev-libs/A-2"},
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/C"],
                success=False,
                options={"--autounmask": True},
                mergelist=["dev-libs/A-9999", "dev-libs/C-1"],
                unstable_keywords={"dev-libs/A-9999"},
                needed_p_mask_changes={"dev-libs/A-9999"},
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds, profile=profile)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 1999-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from itertools import chain
import warnings

import portage
from portage.cache.mappings import slot_dict_class
from portage.const import EBUILD_PHASES
from portage.dep import (
    Atom,
    check_required_use,
    use_reduce,
    paren_enclose,
    _slot_separator,
    _repo_separator,
)
from portage.dep.soname.parse import parse_soname_deps
from portage.versions import _pkg_str, _unknown_repo
from portage.eapi import _get_eapi_attrs
from portage.exception import InvalidData, InvalidDependString
from _emerge.Task import Task


class Package(Task):
    __hash__ = Task.__hash__
    __slots__ = (
        "built",
        "cpv",
        "depth",
        "installed",
        "onlydeps",
        "operation",
        "root_config",
        "type_name",
        "category",
        "counter",
        "cp",
        "cpv_split",
        "inherited",
        "iuse",
        "mtime",
        "pf",
        "root",
        "slot",
        "sub_slot",
        "slot_atom",
        "version",
    ) + (
        "_invalid",
        "_masks",
        "_metadata",
        "_provided_cps",
        "_raw_metadata",
        "_provides",
        "_requires",
        "_use",
        "_validated_atoms",
        "_visible",
    )

    metadata_keys = [
        "BDEPEND",
        "BUILD_ID",
        "BUILD_TIME",
        "CHOST",
        "COUNTER",
        "DEFINED_PHASES",
        "DEPEND",
        "EAPI",
        "IDEPEND",
        "INHERITED",
        "IUSE",
        "KEYWORDS",
        "LICENSE",
        "MD5",
        "PDEPEND",
        "PROVIDES",
        "RDEPEND",
        "repository",
        "REQUIRED_USE",
        "PROPERTIES",
        "REQUIRES",
        "RESTRICT",
        "SIZE",
        "SLOT",
        "USE",
        "_mtime_",
    ]

    _dep_keys = ("BDEPEND", "DEPEND", "IDEPEND", "PDEPEND", "RDEPEND")
    _buildtime_keys = ("BDEPEND", "DEPEND")
    _runtime_keys = ("IDEPEND", "PDEPEND", "RDEPEND")
    _use_conditional_misc_keys = ("LICENSE", "PROPERTIES", "RESTRICT")
    UNKNOWN_REPO = _unknown_repo

    def __init__(self, **kwargs):
        metadata = _PackageMetadataWrapperBase(kwargs.pop("metadata"))
        Task.__init__(self, **kwargs)
        # the SlotObject constructor assigns self.root_config from keyword args
        # and is an instance of a '_emerge.RootConfig.RootConfig class
        self.root = self.root_config.root
        self._raw_metadata = metadata
        self._metadata = _PackageMetadataWrapper(self, metadata)
        if not self.built:
            self._metadata["CHOST"] = self.root_config.settings.get("CHOST", "")
        eapi_attrs = _get_eapi_attrs(self.eapi)

        try:
            db = self.cpv._db
        except AttributeError:
            if self.built:
                # For independence from the source ebuild repository and
                # profile implicit IUSE state, require the _db attribute
                # for built packages.
                raise
            db = self.root_config.trees["porttree"].dbapi

        self.cpv = _pkg_str(
            self.cpv, metadata=self._metadata, settings=self.root_config.settings, db=db
        )
        if hasattr(self.cpv, "slot_invalid"):
            self._invalid_metadata(
                "SLOT.invalid", f"SLOT: invalid value: '{self._metadata['SLOT']}'"
            )
        self.cpv_split = self.cpv.cpv_split
        self.category, self.pf = portage.catsplit(self.cpv)
        self.cp = self.cpv.cp
        self.version = self.cpv.version
        self.slot = self.cpv.slot
        self.sub_slot = self.cpv.sub_slot
        self.slot_atom = Atom(f"{self.cp}{_slot_separator}{self.slot}")
        # sync metadata with validated repo (may be UNKNOWN_REPO)
        self._metadata["repository"] = self.cpv.repo

        implicit_match = db._iuse_implicit_cnstr(self.cpv, self._metadata)
        self.iuse = self._iuse(
            self, self._metadata["IUSE"].split(), implicit_match, self.eapi
        )

        if (self.iuse.enabled or self.iuse.disabled) and not eapi_attrs.iuse_defaults:
            if not self.installed:
                self._invalid_metadata(
                    "EAPI.incompatible",
                    "IUSE contains defaults, but EAPI doesn't allow them",
                )
        if self.inherited is None:
            self.inherited = frozenset()

        if self.operation is None:
            if self.onlydeps or self.installed:
                self.operation = "nomerge"
            else:
                self.operation = "merge"

        self._hash_key = Package._gen_hash_key(
            cpv=self.cpv,
            installed=self.installed,
            onlydeps=self.onlydeps,
            operation=self.operation,
            repo_name=self.cpv.repo,
            root_config=self.root_config,
            type_name=self.type_name,
        )
        self._hash_value = hash(self._hash_key)

    @property
    def eapi(self):
        return self._metadata["EAPI"]

    @property
    def build_id(self):
        return self.cpv.build_id

    @property
    def build_time(self):
        if not self.built:
            raise AttributeError("build_time")
        return self.cpv.build_time

    @property
    def defined_phases(self):
        return self._metadata.defined_phases

    @property
    def properties(self):
        return self._metadata.properties

    @property
    def provided_cps(self):
        return (self.cp,)

    @property
    def restrict(self):
        return self._metadata.restrict

    @property
    def metadata(self):
        warnings.warn(
            "_emerge.Package.Package.metadata is deprecated",
            DeprecationWarning,
            stacklevel=3,
        )
        return self._metadata

    # These are calculated on-demand, so that they are calculated
    # after FakeVartree applies its metadata tweaks.
    @property
    def invalid(self):
        if self._invalid is None:
            self._validate_deps()
            if self._invalid is None:
                self._invalid = False
        return self._invalid

    @property
    def masks(self):
        if self._masks is None:
            self._masks = self._eval_masks()
        return self._masks

    @property
    def visible(self):
        if self._visible is None:
            self._visible = self._eval_visiblity(self.masks)
        return self._visible

    @property
    def validated_atoms(self):
        """
        Returns *all* validated atoms from the deps, regardless
        of USE conditionals, with USE conditionals inside
        atoms left unevaluated.
        """
        if self._validated_atoms is None:
            self._validate_deps()
        return self._validated_atoms

    @property
    def stable(self):
        return self.cpv.stable

    @property
    def provides(self):
        self.invalid
        return self._provides

    @property
    def requires(self):
        self.invalid
        return self._requires

    @classmethod
    def _gen_hash_key(
        cls,
        cpv=None,
        installed=None,
        onlydeps=None,
        operation=None,
        repo_name=None,
        root_config=None,
        type_name=None,
        **kwargs,
    ):
        if operation is None:
            if installed or onlydeps:
                operation = "nomerge"
            else:
                operation = "merge"

        root = None
        if root_config is not None:
            root = root_config.root
        else:
            raise TypeError("root_config argument is required")

        elements = [type_name, root, str(cpv), operation]

        # For installed (and binary) packages we don't care for the repo
        # when it comes to hashing, because there can only be one cpv.
        # So overwrite the repo_key with type_name.
        if type_name is None:
            raise TypeError("type_name argument is required")
        elif type_name == "ebuild":
            if repo_name is None:
                raise AssertionError(
                    "Package._gen_hash_key() " + "called without 'repo_name' argument"
                )
            elements.append(repo_name)
        elif type_name == "binary":
            # Including a variety of fingerprints in the hash makes
            # it possible to simultaneously consider multiple similar
            # packages. Note that digests are not included here, since
            # they are relatively expensive to compute, and they may
            # not necessarily be available.
            elements.extend([cpv.build_id, cpv.file_size, cpv.build_time, cpv.mtime])
        else:
            # For installed (and binary) packages we don't care for the repo
            # when it comes to hashing, because there can only be one cpv.
            # So overwrite the repo_key with type_name.
            elements.append(type_name)

        return tuple(elements)

    def _validate_deps(self):
        """
        Validate deps. This does not trigger USE calculation since that
        is expensive for ebuilds and therefore we want to avoid doing
        it unnecessarily (like for masked packages).
        """
        eapi = self.eapi
        dep_eapi = eapi
        dep_valid_flag = self.iuse.is_valid_flag
        if self.installed:
            # Ignore EAPI.incompatible and conditionals missing
            # from IUSE for installed packages since these issues
            # aren't relevant now (re-evaluate when new EAPIs are
            # deployed).
            dep_eapi = None
            dep_valid_flag = None

        validated_atoms = []
        for k in self._dep_keys:
            v = self._metadata.get(k)
            if not v:
                continue
            try:
                atoms = use_reduce(
                    v,
                    eapi=dep_eapi,
                    matchall=True,
                    is_valid_flag=dep_valid_flag,
                    token_class=Atom,
                    flat=True,
                )
            except InvalidDependString as e:
                self._metadata_exception(k, e)
            else:
                validated_atoms.extend(atoms)
                if not self.built:
                    for atom in atoms:
                        if not isinstance(atom, Atom):
                            continue
                        if atom.slot_operator_built:
                            e = InvalidDependString(
                                'Improper context for slot-operator "built" '
                                f"atom syntax: {atom.unevaluated_atom}"
                            )
                            self._metadata_exception(k, e)

        self._validated_atoms = tuple(
            {atom for atom in validated_atoms if isinstance(atom, Atom)}
        )

        for k in self._use_conditional_misc_keys:
            v = self._metadata.get(k)
            if not v:
                continue
            try:
                use_reduce(
                    v, eapi=dep_eapi, matchall=True, is_valid_flag=dep_valid_flag
                )
            except InvalidDependString as e:
                self._metadata_exception(k, e)

        k = "REQUIRED_USE"
        v = self._metadata.get(k)
        if v and not self.built:
            if not _get_eapi_attrs(eapi).required_use:
                self._invalid_metadata(
                    "EAPI.incompatible",
                    f"REQUIRED_USE set, but EAPI='{eapi}' doesn't allow it",
                )
            else:
                try:
                    check_required_use(v, (), self.iuse.is_valid_flag, eapi=eapi)
                except InvalidDependString as e:
                    self._invalid_metadata(k + ".syntax", f"{k}: {e}")

        k = "SRC_URI"
        v = self._metadata.get(k)
        if v:
            try:
                use_reduce(
                    v,
                    is_src_uri=True,
                    eapi=eapi,
                    matchall=True,
                    is_valid_flag=self.iuse.is_valid_flag,
                )
            except InvalidDependString as e:
                if not self.installed:
                    self._metadata_exception(k, e)

        if self.built:
            k = "PROVIDES"
            try:
                self._provides = frozenset(parse_soname_deps(self._metadata[k]))
            except InvalidData as e:
                self._invalid_metadata(k + ".syntax", f"{k}: {e}")

            k = "REQUIRES"
            try:
                self._requires = frozenset(parse_soname_deps(self._metadata[k]))
            except InvalidData as e:
                self._invalid_metadata(k + ".syntax", f"{k}: {e}")

    def copy(self):
        return Package(
            built=self.built,
            cpv=self.cpv,
            depth=self.depth,
            installed=self.installed,
            metadata=self._raw_metadata,
            onlydeps=self.onlydeps,
            operation=self.operation,
            root_config=self.root_config,
            type_name=self.type_name,
        )

    def _eval_masks(self):
        masks = {}
        settings = self.root_config.settings

        if self.invalid is not False:
            masks["invalid"] = self.invalid

        if not settings._accept_chost(self.cpv, self._metadata):
            masks["CHOST"] = self._metadata["CHOST"]

        eapi = self.eapi
        if not portage.eapi_is_supported(eapi):
            masks["EAPI.unsupported"] = eapi
        if portage._eapi_is_deprecated(eapi):
            masks["EAPI.deprecated"] = eapi

        missing_keywords = settings._getMissingKeywords(self.cpv, self._metadata)
        if missing_keywords:
            masks["KEYWORDS"] = missing_keywords

        try:
            missing_properties = settings._getMissingProperties(
                self.cpv, self._metadata
            )
            if missing_properties:
                masks["PROPERTIES"] = missing_properties
        except InvalidDependString:
            # already recorded as 'invalid'
            pass

        try:
            missing_restricts = settings._getMissingRestrict(self.cpv, self._metadata)
            if missing_restricts:
                masks["RESTRICT"] = missing_restricts
        except InvalidDependString:
            # already recorded as 'invalid'
            pass

        mask_atom = settings._getMaskAtom(self.cpv, self._metadata)
        if mask_atom is not None:
            masks["package.mask"] = mask_atom

        try:
            missing_licenses = settings._getMissingLicenses(self.cpv, self._metadata)
            if missing_licenses:
                masks["LICENSE"] = missing_licenses
        except InvalidDependString:
            # already recorded as 'invalid'
            pass

        if not masks:
            masks = False

        return masks

    def _eval_visiblity(self, masks):
        if masks is not False:
            if "EAPI.unsupported" in masks:
                return False

            if "invalid" in masks:
                return False

            if not self.installed and (
                "CHOST" in masks
                or "EAPI.deprecated" in masks
                or "KEYWORDS" in masks
                or "PROPERTIES" in masks
                or "RESTRICT" in masks
            ):
                return False

            if "package.mask" in masks or "LICENSE" in masks:
                return False

        return True

    def get_keyword_mask(self):
        """returns None, 'missing', or 'unstable'."""

        missing = self.root_config.settings._getRawMissingKeywords(
            self.cpv, self._metadata
        )

        if not missing:
            return None

        if "**" in missing:
            return "missing"

        global_accept_keywords = frozenset(
            self.root_config.settings.get("ACCEPT_KEYWORDS", "").split()
        )

        for keyword in missing:
            if keyword.lstrip("~") in global_accept_keywords:
                return "unstable"

        return "missing"

    def isHardMasked(self):
        """returns a bool if the cpv is in the list of
        expanded pmaskdict[cp] available ebuilds"""
        pmask = self.root_config.settings._getRawMaskAtom(self.cpv, self._metadata)
        return pmask is not None

    def _metadata_exception(self, k, e):
        if k.endswith("DEPEND"):
            qacat = "dependency.syntax"
        else:
            qacat = k + ".syntax"

        if not self.installed:
            categorized_error = False
            if e.errors:
                for error in e.errors:
                    if getattr(error, "category", None) is None:
                        continue
                    categorized_error = True
                    self._invalid_metadata(error.category, f"{k}: {error}")

            if not categorized_error:
                self._invalid_metadata(qacat, f"{k}: {e}")
        else:
            # For installed packages, show the path of the file
            # containing the invalid metadata, since the user may
            # want to fix the deps by hand.
            vardb = self.root_config.trees["vartree"].dbapi
            path = vardb.getpath(self.cpv, filename=k)
            self._invalid_metadata(qacat, f"{k}: {e} in '{path}'")

    def _invalid_metadata(self, msg_type, msg):
        if self._invalid is None:
            self._invalid = {}
        msgs = self._invalid.get(msg_type)
        if msgs is None:
            msgs = []
            self._invalid[msg_type] = msgs
        msgs.append(msg)

    def __str__(self):
        if self.operation == "merge":
            if self.type_name == "binary":
                cpv_color = "PKG_BINARY_MERGE"
            else:
                cpv_color = "PKG_MERGE"
        elif self.operation == "uninstall":
            cpv_color = "PKG_UNINSTALL"
        else:
            cpv_color = "PKG_NOMERGE"

        build_id_str = ""
        if isinstance(self.cpv.build_id, int) and self.cpv.build_id > 0:
            build_id_str = f"-{self.cpv.build_id}"

        s = "({}, {}".format(
            portage.output.colorize(
                cpv_color,
                self.cpv
                + build_id_str
                + _slot_separator
                + self.slot
                + "/"
                + self.sub_slot
                + _repo_separator
                + self.repo,
            ),
            self.type_name,
        )

        if self.type_name == "installed":
            if self.root_config.settings["ROOT"] != "/":
                s += f" in '{self.root_config.settings['ROOT']}'"
            if self.operation == "uninstall":
                s += " scheduled for uninstall"
        else:
            if self.operation == "merge":
                s += " scheduled for merge"
                if self.root_config.settings["ROOT"] != "/":
                    s += f" to '{self.root_config.settings['ROOT']}'"
        s += ")"
        return s

    class _use_class:
        __slots__ = ("enabled", "_expand", "_expand_hidden", "_force", "_pkg", "_mask")

        # Share identical frozenset instances when available.
        _frozensets = {}

        def __init__(self, pkg, enabled_flags):
            self._pkg = pkg
            self._expand = None
            self._expand_hidden = None
            self._force = None
            self._mask = None
            self.enabled = frozenset(enabled_flags)
            if pkg.built:
                # Use IUSE to validate USE settings for built packages,
                # in case the package manager that built this package
                # failed to do that for some reason (or in case of
                # data corruption).
                missing_iuse = pkg.iuse.get_missing_iuse(self.enabled)
                if missing_iuse:
                    self.enabled = self.enabled.difference(missing_iuse)

        def _init_force_mask(self):
            pkgsettings = self._pkg._get_pkgsettings()
            frozensets = self._frozensets
            s = frozenset(pkgsettings.get("USE_EXPAND", "").lower().split())
            self._expand = frozensets.setdefault(s, s)
            s = frozenset(pkgsettings.get("USE_EXPAND_HIDDEN", "").lower().split())
            self._expand_hidden = frozensets.setdefault(s, s)
            s = pkgsettings.useforce
            self._force = frozensets.setdefault(s, s)
            s = pkgsettings.usemask
            self._mask = frozensets.setdefault(s, s)

        @property
        def expand(self):
            if self._expand is None:
                self._init_force_mask()
            return self._expand

        @property
        def expand_hidden(self):
            if self._expand_hidden is None:
                self._init_force_mask()
            return self._expand_hidden

        @property
        def force(self):
            if self._force is None:
                self._init_force_mask()
            return self._force

        @property
        def mask(self):
            if self._mask is None:
                self._init_force_mask()
            return self._mask

    @property
    def repo(self):
        return self._metadata["repository"]

    @property
    def repo_priority(self):
        repo_info = self.root_config.settings.repositories.prepos.get(self.repo)
        if repo_info is None:
            return None
        return repo_info.priority

    @property
    def use(self):
        if self._use is None:
            self._init_use()
        return self._use

    def _get_pkgsettings(self):
        pkgsettings = self.root_config.trees["porttree"].dbapi.doebuild_settings
        pkgsettings.setcpv(self)
        return pkgsettings

    def _init_use(self):
        if self.built:
            # Use IUSE to validate USE settings for built packages,
            # in case the package manager that built this package
            # failed to do that for some reason (or in case of
            # data corruption). The enabled flags must be consistent
            # with implicit IUSE, in order to avoid potential
            # inconsistencies in USE dep matching (see bug #453400).
            use_str = self._metadata["USE"]
            is_valid_flag = self.iuse.is_valid_flag
            enabled_flags = [x for x in use_str.split() if is_valid_flag(x)]
            use_str = " ".join(enabled_flags)
            self._use = self._use_class(self, enabled_flags)
        else:
            try:
                use_str = _PackageMetadataWrapperBase.__getitem__(self._metadata, "USE")
            except KeyError:
                use_str = None
            calculated_use = False
            if not use_str:
                use_str = self._get_pkgsettings()["PORTAGE_USE"]
                calculated_use = True
            self._use = self._use_class(self, use_str.split())
            # Initialize these now, since USE access has just triggered
            # setcpv, and we want to cache the result of the force/mask
            # calculations that were done.
            if calculated_use:
                self._use._init_force_mask()

        _PackageMetadataWrapperBase.__setitem__(self._metadata, "USE", use_str)

        return use_str

    class _iuse:
        __slots__ = (
            "__weakref__",
            "_iuse_implicit_match",
            "_pkg",
            "all",
            "enabled",
            "disabled",
            "tokens",
        )

        def __init__(self, pkg, tokens, iuse_implicit_match, eapi):
            self._pkg = pkg
            self.tokens = tuple(tokens)
            self._iuse_implicit_match = iuse_implicit_match
            enabled = []
            disabled = []
            other = []
            for x in tokens:
                prefix = x[:1]
                if prefix == "+":
                    enabled.append(x[1:])
                elif prefix == "-":
                    disabled.append(x[1:])
                else:
                    other.append(x)
            self.enabled = frozenset(enabled)
            self.disabled = frozenset(disabled)
            self.all = frozenset(chain(enabled, disabled, other))

        def is_valid_flag(self, flags):
            """
            @return: True if all flags are valid USE values which may
                    be specified in USE dependencies, False otherwise.
            """
            if isinstance(flags, str):
                flags = [flags]

            for flag in flags:
                if self.get_flag(flag) is None:
                    return False
            return True

        def get_missing_iuse(self, flags):
            """
            @return: A list of flags missing from IUSE.
            """
            if isinstance(flags, str):
                flags = [flags]
            missing_iuse = []
            for flag in flags:
                if self.get_flag(flag) is None:
                    missing_iuse.append(flag)
            return missing_iuse

        def get_flag(self, flag):
            """
            Returns the flag's name within the scope of this package
            or None if the flag is unknown.
            """
            if flag in self.all:
                return flag

            if self._iuse_implicit_match(flag):
                return flag

            return None

    def __len__(self):
        return 4

    def __iter__(self):
        """
        This is used to generate mtimedb resume mergelist entries, so we
        limit it to 4 items for backward compatibility.
        """
        return iter(self._hash_key[:4])

    def __lt__(self, other):
        if other.cp != self.cp:
            return self.cp < other.cp
        result = portage.vercmp(self.version, other.version)
        if result < 0:
            return True
        if result == 0 and self.built and other.built:
            return self.build_time < other.build_time
        return False

    def __le__(self, other):
        if other.cp != self.cp:
            return self.cp <= other.cp
        result = portage.vercmp(self.version, other.version)
        if result <= 0:
            return True
        if result == 0 and self.built and other.built:
            return self.build_time <= other.build_time
        return False

    def __gt__(self, other):
        if other.cp != self.cp:
            return self.cp > other.cp
        result = portage.vercmp(self.version, other.version)
        if result > 0:
            return True
        if result == 0 and self.built and other.built:
            return self.build_time > other.build_time
        return False

    def __ge__(self, other):
        if other.cp != self.cp:
            return self.cp >= other.cp
        result = portage.vercmp(self.version, other.version)
        if result >= 0:
            return True
        if result == 0 and self.built and other.built:
            return self.build_time >= other.build_time
        return False

    def with_use(self, use):
        """
        Return an Package instance with the specified USE flags. The
        current instance may be returned if it has identical USE flags.
        @param use: a set of USE flags
        @type use: frozenset
        @return: A package with the specified USE flags
        @rtype: Package
        """
        if use is not self.use.enabled:
            pkg = self.copy()
            pkg._metadata["USE"] = " ".join(use)
        else:
            pkg = self
        return pkg


_all_metadata_keys = frozenset(set(portage.auxdbkeys).union(Package.metadata_keys))

_PackageMetadataWrapperBase = slot_dict_class(_all_metadata_keys)


class _PackageMetadataWrapper(_PackageMetadataWrapperBase):
    """
    Detect metadata updates and synchronize Package attributes.
    """

    __slots__ = ("_pkg",)
    _wrapped_keys = frozenset(["COUNTER", "INHERITED", "USE", "_mtime_"])
    _use_conditional_keys = frozenset(
        [
            "LICENSE",
            "PROPERTIES",
            "RESTRICT",
        ]
    )

    def __init__(self, pkg, metadata):
        _PackageMetadataWrapperBase.__init__(self)
        self._pkg = pkg
        if not pkg.built:
            # USE is lazy, but we want it to show up in self.keys().
            _PackageMetadataWrapperBase.__setitem__(self, "USE", "")

        self.update(metadata)

    def __getitem__(self, k):
        v = _PackageMetadataWrapperBase.__getitem__(self, k)
        if k in self._use_conditional_keys:
            if self._pkg.root_config.settings.local_config and "?" in v:
                try:
                    v = paren_enclose(
                        use_reduce(
                            v,
                            uselist=self._pkg.use.enabled,
                            is_valid_flag=self._pkg.iuse.is_valid_flag,
                        )
                    )
                except InvalidDependString:
                    # This error should already have been registered via
                    # self._pkg._invalid_metadata().
                    pass
                else:
                    self[k] = v

        elif k == "USE" and not self._pkg.built:
            if not v:
                # This is lazy because it's expensive.
                v = self._pkg._init_use()

        return v

    def __setitem__(self, k, v):
        _PackageMetadataWrapperBase.__setitem__(self, k, v)
        if k in self._wrapped_keys:
            getattr(self, "_set_" + k.lower())(k, v)

    def _set_inherited(self, k, v):
        if isinstance(v, str):
            v = frozenset(v.split())
        self._pkg.inherited = v

    def _set_counter(self, k, v):
        if isinstance(v, str):
            try:
                v = int(v.strip())
            except ValueError:
                v = 0
        self._pkg.counter = v

    def _set_use(self, k, v):
        # Force regeneration of _use attribute
        self._pkg._use = None
        # Use raw metadata to restore USE conditional values
        # to unevaluated state
        raw_metadata = self._pkg._raw_metadata
        for x in self._use_conditional_keys:
            try:
                self[x] = raw_metadata[x]
            except KeyError:
                pass

    def _set__mtime_(self, k, v):
        if isinstance(v, str):
            try:
                v = int(v.strip())
            except ValueError:
                v = 0
        self._pkg.mtime = v

    @property
    def properties(self):
        return self["PROPERTIES"].split()

    @property
    def restrict(self):
        return self["RESTRICT"].split()

    @property
    def defined_phases(self):
        """
        Returns tokens from DEFINED_PHASES metadata if it is defined,
        otherwise returns a tuple containing all possible phases. This
        makes it easy to do containment checks to see if it's safe to
        skip execution of a given phase.
        """
        s = self["DEFINED_PHASES"]
        if s:
            return s.split()
        return EBUILD_PHASES



# Copyright 2012-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from _emerge.SubProcess import SubProcess


class PopenProcess(SubProcess):
    __slots__ = (
        "pipe_reader",
        "proc",
    )

    def _start(self):
        self.pid = self.proc.pid
        self._registered = True

        if self.pipe_reader is None:
            self.scheduler.call_soon(self._async_waitpid)
        else:
            try:
                self.pipe_reader.scheduler = self.scheduler
            except AttributeError:
                pass
            self.pipe_reader.addExitListener(self._pipe_reader_exit)
            self.pipe_reader.start()

    def _pipe_reader_exit(self, pipe_reader):
        self._async_waitpid()

    def _async_waitpid_cb(self, *args, **kwargs):
        SubProcess._async_waitpid_cb(self, *args, **kwargs)
        if self.proc.returncode is None:
            # Suppress warning messages like this:
            # ResourceWarning: subprocess 1234 is still running
            self.proc.returncode = self.returncode



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ["movefile"]

import errno
import fnmatch
import os as _os
import stat
import textwrap

import portage
from portage import (
    bsd_chflags,
    _encodings,
    _os_overrides,
    _selinux,
    _unicode_decode,
    _unicode_encode,
    _unicode_func_wrapper,
    _unicode_module_wrapper,
)
from portage.const import MOVE_BINARY
from portage.exception import OperationNotSupported
from portage.localization import _
from portage.process import spawn
from portage.util import writemsg
from portage.util._xattr import xattr
from portage.util.file_copy import copyfile


def _apply_stat(src_stat, dest):
    _os.chown(dest, src_stat.st_uid, src_stat.st_gid)
    _os.chmod(dest, stat.S_IMODE(src_stat.st_mode))


_xattr_excluder_cache = {}


def _get_xattr_excluder(pattern):
    try:
        value = _xattr_excluder_cache[pattern]
    except KeyError:
        value = _xattr_excluder(pattern)
        _xattr_excluder_cache[pattern] = value

    return value


class _xattr_excluder:
    __slots__ = ("_pattern_split",)

    def __init__(self, pattern):
        if pattern is None:
            self._pattern_split = None
        else:
            pattern = pattern.split()
            if not pattern:
                self._pattern_split = None
            else:
                pattern.sort()
                self._pattern_split = tuple(pattern)

    def __call__(self, attr):
        if self._pattern_split is None:
            return False

        match = fnmatch.fnmatch
        for x in self._pattern_split:
            if match(attr, x):
                return True

        return False


def _copyxattr(src, dest, exclude=None):
    """Copy the extended attributes from |src| to |dest|"""
    try:
        attrs = xattr.list(src)
    except OSError as e:
        if e.errno != OperationNotSupported.errno:
            raise
        attrs = ()

    if attrs:
        if exclude is not None and isinstance(attrs[0], bytes):
            exclude = exclude.encode(_encodings["fs"])
        exclude = _get_xattr_excluder(exclude)

    for attr in attrs:
        if exclude(attr):
            continue
        try:
            xattr.set(dest, attr, xattr.get(src, attr))
            raise_exception = False
        except OSError:
            raise_exception = True
        if raise_exception:
            raise OperationNotSupported(
                _(
                    "Filesystem containing file '%s' "
                    "does not support extended attribute '%s'"
                )
                % (_unicode_decode(dest), _unicode_decode(attr))
            )


def _cmpxattr(src, dest, exclude=None):
    """
    Compares extended attributes between |src| and |dest| and returns True
    if they are equal or xattrs are not supported, False otherwise
    """
    try:
        src_attrs = xattr.list(src)
        dest_attrs = xattr.list(dest)
    except OSError as e:
        if e.errno != OperationNotSupported.errno:
            raise
        return True

    if src_attrs:
        if exclude is not None and isinstance(src_attrs[0], bytes):
            exclude = exclude.encode(_encodings["fs"])
    exclude = _get_xattr_excluder(exclude)

    src_attrs = {attr for attr in src_attrs if not exclude(attr)}
    dest_attrs = {attr for attr in dest_attrs if not exclude(attr)}
    if src_attrs != dest_attrs:
        return False

    for attr in src_attrs:
        if xattr.get(src, attr) != xattr.get(dest, attr):
            return False
    return True


def movefile(
    src,
    dest,
    newmtime=None,
    sstat=None,
    mysettings=None,
    hardlink_candidates=None,
    encoding=_encodings["fs"],
):
    """moves a file from src to dest, preserving all permissions and attributes; mtime will
    be preserved even when moving across filesystems.  Returns mtime as integer on success
    and None on failure.  mtime is expressed in seconds in Python <3.3 and nanoseconds in
    Python >=3.3.  Move is atomic."""

    if mysettings is None:
        mysettings = portage.settings

    src_bytes = _unicode_encode(src, encoding=encoding, errors="strict")
    dest_bytes = _unicode_encode(dest, encoding=encoding, errors="strict")
    xattr_enabled = "xattr" in mysettings.features
    selinux_enabled = mysettings.selinux_enabled()
    if selinux_enabled:
        selinux = _unicode_module_wrapper(_selinux, encoding=encoding)
        _copyfile = selinux.copyfile
        _rename = selinux.rename
    else:
        _copyfile = copyfile
        _rename = _os.rename

    lchown = _unicode_func_wrapper(portage.data.lchown, encoding=encoding)
    os = _unicode_module_wrapper(_os, encoding=encoding, overrides=_os_overrides)

    try:
        if not sstat:
            sstat = os.lstat(src)

    except SystemExit as e:
        raise
    except Exception as e:
        writemsg(
            f"!!! {_('Stating source file failed... movefile()')}\n", noiselevel=-1
        )
        writemsg(f"!!! {e}\n", noiselevel=-1)
        return None

    destexists = 1
    try:
        dstat = os.lstat(dest)
    except OSError:
        dstat = os.lstat(os.path.dirname(dest))
        destexists = 0

    if bsd_chflags:
        if destexists and dstat.st_flags != 0:
            bsd_chflags.lchflags(dest, 0)
        # Use normal stat/chflags for the parent since we want to
        # follow any symlinks to the real parent directory.
        pflags = os.stat(os.path.dirname(dest)).st_flags
        if pflags != 0:
            bsd_chflags.chflags(os.path.dirname(dest), 0)

    if destexists:
        if not stat.S_ISLNK(sstat[stat.ST_MODE]) and stat.S_ISLNK(dstat[stat.ST_MODE]):
            try:
                os.unlink(dest)
                destexists = 0
            except SystemExit as e:
                raise
            except Exception as e:
                pass

    if stat.S_ISLNK(sstat[stat.ST_MODE]):
        try:
            target = os.readlink(src)
            if mysettings and "D" in mysettings and target.startswith(mysettings["D"]):
                target = target[len(mysettings["D"]) - 1 :]
            # Atomically update the path if it exists.
            try:
                os.rename(src, dest)
                return sstat.st_mtime_ns
            except OSError:
                # If it failed due to cross-link device, fallthru below.
                # Clear the target first so we can create it.
                try:
                    os.unlink(dest)
                except FileNotFoundError:
                    pass

            try:
                if selinux_enabled:
                    selinux.symlink(target, dest, src)
                else:
                    os.symlink(target, dest)
            except OSError as e:
                # Some programs will create symlinks automatically, so we have
                # to tolerate these links being recreated during the merge
                # process. In any case, if the link is pointing at the right
                # place, we're in good shape.
                if e.errno not in (errno.ENOENT, errno.EEXIST) or target != os.readlink(
                    dest
                ):
                    raise
            lchown(dest, sstat[stat.ST_UID], sstat[stat.ST_GID])

            try:
                _os.unlink(src_bytes)
            except OSError:
                pass

            try:
                os.utime(
                    dest,
                    ns=(sstat.st_mtime_ns, sstat.st_mtime_ns),
                    follow_symlinks=False,
                )
            except NotImplementedError:
                # utimensat() and lutimes() missing in libc.
                return os.stat(dest, follow_symlinks=False).st_mtime_ns
            else:
                return sstat.st_mtime_ns
        except SystemExit as e:
            raise
        except Exception as e:
            writemsg(f"!!! {_('failed to properly create symlink:')}\n", noiselevel=-1)
            writemsg(f"!!! {dest} -> {target}\n", noiselevel=-1)
            writemsg(f"!!! {e}\n", noiselevel=-1)
            return None

    hardlinked = False
    # Since identical files might be merged to multiple filesystems,
    # so os.link() calls might fail for some paths, so try them all.
    # For atomic replacement, first create the link as a temp file
    # and them use os.rename() to replace the destination.
    if hardlink_candidates:
        head, tail = os.path.split(dest)
        hardlink_tmp = os.path.join(head, f".{tail}._portage_merge_.{portage.getpid()}")
        try:
            os.unlink(hardlink_tmp)
        except OSError as e:
            if e.errno != errno.ENOENT:
                writemsg(
                    _("!!! Failed to remove hardlink temp file: %s\n")
                    % (hardlink_tmp,),
                    noiselevel=-1,
                )
                writemsg(f"!!! {e}\n", noiselevel=-1)
                return None
            del e
        for hardlink_src in hardlink_candidates:
            try:
                os.link(hardlink_src, hardlink_tmp)
            except OSError:
                continue
            else:
                try:
                    os.rename(hardlink_tmp, dest)
                except OSError as e:
                    writemsg(
                        _("!!! Failed to rename %s to %s\n") % (hardlink_tmp, dest),
                        noiselevel=-1,
                    )
                    writemsg(f"!!! {e}\n", noiselevel=-1)
                    return None
                hardlinked = True
                try:
                    _os.unlink(src_bytes)
                except OSError:
                    pass
                break

    renamefailed = 1
    if hardlinked:
        renamefailed = False
    if not hardlinked and (selinux_enabled or sstat.st_dev == dstat.st_dev):
        try:
            if selinux_enabled:
                selinux.rename(src, dest)
            else:
                os.rename(src, dest)
            renamefailed = 0
        except OSError as e:
            if e.errno != errno.EXDEV:
                # Some random error.
                writemsg(
                    f"!!! {_('Failed to move %(src)s to %(dest)s')}\n"
                    % {"src": src, "dest": dest},
                    noiselevel=-1,
                )
                writemsg(f"!!! {e}\n", noiselevel=-1)
                return None
            # Invalid cross-device-link 'bind' mounted or actually Cross-Device
    if renamefailed:
        if stat.S_ISREG(sstat[stat.ST_MODE]):
            dest_tmp = dest + "#new"
            dest_tmp_bytes = _unicode_encode(
                dest_tmp, encoding=encoding, errors="strict"
            )
            success = False
            try:  # For safety copy then move it over.
                _copyfile(src_bytes, dest_tmp_bytes)
                _apply_stat(sstat, dest_tmp_bytes)
                if xattr_enabled:
                    try:
                        _copyxattr(
                            src_bytes,
                            dest_tmp_bytes,
                            exclude=mysettings.get("PORTAGE_XATTR_EXCLUDE", ""),
                        )
                    except SystemExit:
                        raise
                    except:
                        msg = _(
                            "Failed to copy extended attributes. "
                            "In order to avoid this error, set "
                            'FEATURES="-xattr" in make.conf.'
                        )
                        msg = textwrap.wrap(msg, 65)
                        for line in msg:
                            writemsg(f"!!! {line}\n", noiselevel=-1)
                        raise
                _rename(dest_tmp_bytes, dest_bytes)
                _os.unlink(src_bytes)
                success = True
            except Exception as e:
                writemsg(
                    f"!!! {_('copy %(src)s -> %(dest)s failed.')}\n"
                    % {"src": src, "dest": dest},
                    noiselevel=-1,
                )
                writemsg(f"!!! {e}\n", noiselevel=-1)
                return None
            finally:
                if not success:
                    try:
                        _os.unlink(dest_tmp_bytes)
                    except OSError:
                        pass
        else:
            # we don't yet handle special, so we need to fall back to /bin/mv
            a = spawn([MOVE_BINARY, "-f", src, dest], env=os.environ)
            if a != os.EX_OK:
                writemsg(_("!!! Failed to move special file:\n"), noiselevel=-1)
                writemsg(
                    _("!!! '%(src)s' to '%(dest)s'\n")
                    % {
                        "src": _unicode_decode(src, encoding=encoding),
                        "dest": _unicode_decode(dest, encoding=encoding),
                    },
                    noiselevel=-1,
                )
                writemsg(f"!!! {a}\n", noiselevel=-1)
                return None  # failure

    # In Python <3.3 always use stat_obj[stat.ST_MTIME] for the integral timestamp
    # which is returned, since the stat_obj.st_mtime float attribute rounds *up*
    # if the nanosecond part of the timestamp is 999999881 ns or greater.
    try:
        if hardlinked:
            newmtime = os.stat(dest).st_mtime_ns
        else:
            # Note: It is not possible to preserve nanosecond precision
            # (supported in POSIX.1-2008 via utimensat) with the IEEE 754
            # double precision float which only has a 53 bit significand.
            if newmtime is not None:
                os.utime(dest, ns=(newmtime, newmtime))
            else:
                newmtime = sstat.st_mtime_ns
                if renamefailed:
                    # If rename succeeded then timestamps are automatically
                    # preserved with complete precision because the source
                    # and destination inodes are the same. Otherwise, manually
                    # update timestamps with nanosecond precision.
                    os.utime(dest, ns=(newmtime, newmtime))
    except OSError:
        # The utime can fail here with EPERM even though the move succeeded.
        # Instead of failing, use stat to return the mtime if possible.
        try:
            newmtime = os.stat(dest).st_mtime_ns
        except OSError as e:
            writemsg(_("!!! Failed to stat in movefile()\n"), noiselevel=-1)
            writemsg(f"!!! {dest}\n", noiselevel=-1)
            writemsg(f"!!! {str(e)}\n", noiselevel=-1)
            return None

    if bsd_chflags:
        # Restore the flags we saved before moving
        if pflags:
            bsd_chflags.chflags(os.path.dirname(dest), pflags)

    return newmtime



# Copyright 2018-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

doc = """Mercurial plug-in module for portage.
Performs a hg pull on repositories."""
__doc__ = doc[:]

from portage.localization import _
from portage.sync.config_checks import CheckSyncConfig
from portage.util import writemsg_level

module_spec = {
    "name": "mercurial",
    "description": doc,
    "provides": {
        "mercurial-module": {
            "name": "mercurial",
            "sourcefile": "mercurial",
            "class": "MercurialSync",
            "description": doc,
            "functions": ["sync", "new", "exists", "retrieve_head"],
            "func_desc": {
                "sync": "Performs a hg pull on the repository",
                "new": "Creates the new repository at the specified location",
                "exists": "Returns a boolean of whether the specified dir "
                + "exists and is a valid Mercurial repository",
                "retrieve_head": "Returns the head commit hash",
            },
            "validate_config": CheckSyncConfig,
            "module_specific_options": (
                "sync-mercurial-clone-env",
                "sync-mercurial-clone-extra-opts",
                "sync-mercurial-env",
                "sync-mercurial-pull-env",
                "sync-mercurial-pull-extra-opts",
            ),
        }
    },
}
